"""this module is implement to deal the tracker all data

"""
import copy

import logging
import threading
from tingyun.logistics.workshop.packets import TimePackets, ApdexPackets, SlowSqlPackets
from tingyun.packages import six


console = logging.getLogger(__name__)


class Packager(object):
    """
    """
    def __init__(self):
        self.__max_error_count = 10
        self.__settings = None

        self.__time_packets = {}  # store time packets data key is name + scope
        self.__apdex_packets = {}
        self.__action_packets = {}
        self.__general_packets = {}
        # it's maybe include {name: {trackerType, count, $ERROR_ITEM}}
        # $ERROR_ITEM  detail, check the documentation
        self.__traced_errors = {}

        # it's maybe include {name: {count, $ERROR_ITEM}}
        # $ERROR_ITEM  detail, check the documentation
        self.__traced_external_errors = {}

        # format: {$"metric_name": [$slow_action_data, $duration]}
        self.__slow_action = {}
        self.__slow_sql_packets = {}

        self._packets_lock = threading.Lock()

    @property
    def settings(self):
        """
        :return:
        """
        return self.__settings

    def create_data_zone(self):
        """create empty data engine to contain the unmerged tracker data
        :return:
        """
        zone = Packager()
        zone.__settings = self.__settings

        return zone

    def reset_packets(self, application_settings):
        """
        :param application_settings: application settings from server and setting file
        :return:
        """
        self.__settings = application_settings

        self.__time_packets = {}
        self.__apdex_packets = {}
        self.__action_packets = {}
        self.__general_packets = {}
        self.__traced_errors = {}
        self.__traced_external_errors = {}
        self.__slow_action = {}
        self.__slow_sql_packets = {}

    def record_tracker(self, tracker):
        """
        :param tracker: tracker node instance
        :return:
        """
        if not self.__settings:
            console.error("The application settings is not merge into data engine.")
            return

        node_limit = self.__settings.action_tracer_nodes
        threshold = self.__settings.action_tracer.action_threshold

        self.record_time_metrics(tracker.time_metrics())  # deal for component, include framework/db/other
        self.record_action_metrics(tracker.action_metrics())  # deal for user action
        self.record_apdex_metrics(tracker.apdex_metrics())  # for recording the apdex
        self.record_traced_errors(tracker.traced_error())  # for error trace detail
        self.record_slow_action(tracker.slow_action_trace(node_limit, threshold))
        self.record_slow_sql(tracker.slow_sql_nodes())
        self.record_traced_external_trace(tracker.traced_external_error())

    def record_slow_sql(self, nodes):
        """
        :param nodes: the slow sql node
        :return:
        """
        if not self.__settings.action_tracer.slow_sql:
            return

        for node in nodes:
            key = node.identifier
            packets = self.__slow_sql_packets.get(key)
            if not packets and len(self.__slow_sql_packets) < self.__settings.slow_sql_count:
                packets = SlowSqlPackets()
                self.__slow_sql_packets[key] = packets

            if packets:
                packets.merge_slow_sql_node(node)

    def record_slow_action(self, slow_action):
        """
        :param slow_action:
        :return:
        """
        if not self.__settings.action_tracer.enabled:
            return

        threshold = self.__settings.action_tracer.action_threshold
        if threshold > slow_action[1]:
            return

        top_n = self.__settings.action_tracer.top_n
        metric_name = slow_action[2]

        if metric_name not in self.__slow_action:
            self.__slow_action[metric_name] = [[slow_action, slow_action[1]]]
        else:
            # every request url/metric can save top_n action trace.
            if len(self.__slow_action[metric_name]) <= top_n:
                self.__slow_action[metric_name].append([slow_action, slow_action[1]])

    def record_traced_errors(self, traced_errors):
        """
        :return:
        """
        if not self.__settings.error_collector.enabled:
            return

        for error in traced_errors:
            if len(self.__traced_errors) > self.__max_error_count:
                console.debug("Error trace is reached maximum limitation.")
                break

            if error.error_filter_key in self.__traced_errors:
                self.__traced_errors[error.error_filter_key]["count"] += 1
                self.__traced_errors[error.error_filter_key]["item"][-3] += 1
            else:
                self.__traced_errors[error.error_filter_key] = {"count": 1,
                                                                "item": error.trace_data,
                                                                "tracker_type": error.tracker_type}

    def record_traced_external_trace(self, traced_errors):
        """
        :param traced_errors:
        :return:
        """
        if not self.__settings.error_collector.enabled:
            return

        for error in traced_errors:
            if len(self.__traced_external_errors) > self.__max_error_count:
                console.debug("External error trace is reached maximum limitation.")
                break

            if error.error_filter_key in self.__traced_external_errors:
                self.__traced_external_errors[error.error_filter_key]["count"] += 1
                self.__traced_external_errors[error.error_filter_key]["item"][-3] += 1

            else:
                self.__traced_external_errors[error.error_filter_key] = {"count": 1, "status_code": error.status_code,
                                                                         "item": error.trace_data,
                                                                         "tracker_type": error.tracker_type}

    def record_general_metric(self, metric):
        """
        :param metric:
        :return:
        """
        key = (metric.name.split("/", 1)[1], '')

        packets = self.__general_packets.get(key)
        if packets is None:
            packets = TimePackets()

        packets.merge_time_metric(metric)
        self.__general_packets[key] = packets

        return key

    def record_time_metric(self, metric):
        """
        :param metric:
        :return:
        """
        # filter the general data from the metric, the metric node should be distinguish general and basic metric
        if metric.name.startswith("GENERAL"):
            self.record_general_metric(metric)
            return

        key = (metric.name, metric.scope or '')  # metric key for protocol
        packets = self.__time_packets.get(key)
        if packets is None:
            packets = TimePackets()

        packets.merge_time_metric(metric)
        self.__time_packets[key] = packets

        return key

    def record_time_metrics(self, metrics):
        """
        :param metrics:
        :return:
        """
        for metric in metrics:
            self.record_time_metric(metric)

    def record_apdex_metric(self, metric):
        """
        :param metric:
        :return:
        """
        key = (metric.name, "")
        packets = self.__apdex_packets.get(key)

        if packets is None:
            packets = ApdexPackets(apdex_t=metric.apdex_t)

        packets.merge_apdex_metric(metric)
        self.__apdex_packets[key] = packets
        return key

    def record_apdex_metrics(self, metrics):
        """
        :param metrics:
        :return:
        """
        for metric in metrics:
            self.record_apdex_metric(metric)

    def record_action_metric(self, metric):
        """
        :param metric:
        :return:
        """
        key = (metric.name, metric.scope or '')  # metric key for protocol
        packets = self.__action_packets.get(key)
        if packets is None:
            packets = TimePackets()

        packets.merge_time_metric(metric)
        self.__action_packets[key] = packets

        return key

    def record_action_metrics(self, metrics):
        """
        :param metrics:
        :return:
        """
        for metric in metrics:
            self.record_action_metric(metric)

    def rollback(self, stat, merge_performance=True):
        """rollback the performance data when upload the data failed. except the traced error count.
        :param stat:
        :param merge_performance:
        :return:
        """
        if not merge_performance:
            return

        console.warning("Agent will rollback the data which is captured at last time. That indicates your network is"
                        " broken.")

        for key, value in six.iteritems(stat.__time_packets):
            packets = self.__time_packets.get(key)
            if not packets:
                self.__time_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        for key, value in six.iteritems(stat.__apdex_packets):
            packets = self.__apdex_packets.get(key)
            if not packets:
                self.__apdex_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        for key, value in six.iteritems(stat.__action_packets):
            packets = self.__action_packets.get(key)
            if not packets:
                self.__action_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        for key, value in six.iteritems(stat.__general_packets):
            packets = self.__general_packets.get(key)
            if not packets:
                self.__general_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        for key, value in six.iteritems(stat.__traced_errors):
            packets = self.__traced_errors.get(key)
            if not packets:
                self.__traced_errors[key] = copy.copy(value)
            else:
                packets["count"] += value["count"]

        for key, value in six.iteritems(stat.__traced_external_errors):
            packets = self.__traced_external_errors.get(key)
            if not packets:
                self.__traced_external_errors[key] = copy.copy(value)
            else:
                packets["count"] += value["count"]

    def merge_metric_packets(self, snapshot):
        """
        :param snapshot:
        :return:
        """
        for key, value in six.iteritems(snapshot.__time_packets):
            packets = self.__time_packets.get(key)
            if not packets:
                self.__time_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        for key, value in six.iteritems(snapshot.__apdex_packets):
            packets = self.__apdex_packets.get(key)
            if not packets:
                self.__apdex_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        for key, value in six.iteritems(snapshot.__action_packets):
            packets = self.__action_packets.get(key)
            if not packets:
                self.__action_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        # TODO: think more about the background task
        for key, value in six.iteritems(snapshot.__traced_errors):
            packets = self.__traced_errors.get(key)
            if not packets:
                self.__traced_errors[key] = copy.copy(value)
            else:
                packets["item"][-3] += value["count"]

        for key, value in six.iteritems(snapshot.__traced_external_errors):
            packets = self.__traced_external_errors.get(key)
            if not packets:
                self.__traced_external_errors[key] = copy.copy(value)
            else:
                packets["item"][-3] += value["count"]

        # generate general data
        for key, value in six.iteritems(snapshot.__general_packets):
            packets = self.__general_packets.get(key)
            if not packets:
                self.__general_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        # for action trace
        top_n = self.__settings.action_tracer.top_n
        for key, value in six.iteritems(snapshot.__slow_action):

            if key not in self.__slow_action:
                self.__slow_action[key] = value
                break

            slow_actons = self.__slow_action.get(key)
            # although the target action trace value is `list`, but it only has 1 element in one metric.
            if len(slow_actons) > top_n:
                console.debug("The action trace is reach the top(%s), action(%s) is ignored.", top_n, key)
                break
            slow_actons.extend(value)

        # for slow sql
        max_sql = self.__settings.slow_sql_count
        for key, value in six.iteritems(snapshot.__slow_sql_packets):
            if len(self.__slow_sql_packets) > max_sql:
                console.debug("The slow sql trace count is reach the top.")

            slow_sql = self.__slow_sql_packets.get(key)
            if not slow_sql:
                self.__slow_sql_packets[key] = value
            else:
                if value.slow_sql_node.duration > slow_sql.slow_sql_node.duration:
                    self.__slow_sql_packets[key] = value

    def reset_metric_packets(self):
        """
        :return:
        """
        self.__time_packets = {}  # component
        self.__apdex_packets = {}
        self.__action_packets = {}
        self.__general_packets = {}

        self.__traced_errors = {}
        self.__traced_external_errors = {}
        self.__slow_action = {}
        self.__slow_sql_packets = {}

    def packets_snapshot(self):
        """
        :return:
        """
        stat = copy.copy(self)

        self.__time_packets = {}
        self.__action_packets = {}
        self.__apdex_packets = {}
        self.__traced_errors = {}
        self.__traced_external_errors = {}
        self.__general_packets = {}
        self.__slow_action = {}
        self.__slow_sql_packets = {}

        return stat

    # just for upload performance package
    # strip to 5 parts
    def component_metrics(self, metric_name_ids):
        """
        :return:
        """
        result = []

        for key, value in six.iteritems(self.__time_packets):
            extend_metrics = key[0].split("|")
            if len(extend_metrics) == 1:
                upload_key = {"name": key[0], "parent": key[1]}
                upload_key_str = '%s:%s' % (key[0], key[1])
                upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]
                result.append([upload_key, value])
            elif len(extend_metrics) == 3:
                upload_key = {"name": extend_metrics[0], "parent": key[1], "calleeId": extend_metrics[1],
                              "calleeName": extend_metrics[2]}
                result.append([upload_key, value])

        return result

    def apdex_data(self, metric_name_ids):
        """
        :return:
        """
        result = []

        for key, value in six.iteritems(self.__apdex_packets):
            upload_key = {"name": key[0]}
            upload_key_str = '%s' % key[0]
            upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]
            result.append([upload_key, value])

        return result

    def action_metrics(self, metric_name_ids):
        """
        :return:
        """
        result = []
        for key, value in six.iteritems(self.__action_packets):
            upload_key = {"name": key[0]}
            upload_key_str = '%s' % key[0]
            upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]
            result.append([upload_key, value])

        return result

    def error_packets(self, metric_name_ids):
        """stat the error trace metric for performance
        :return:
        """
        external_error, web_action_error = 'External', 'WebAction'
        error_types = [external_error, web_action_error]
        error_count = {
            "Errors/Count/All": 0,
            "Errors/Count/AllWeb": 0,
            "Errors/Count/AllBackground": 0
        }

        def parse_error_trace(traced_data):
            for error_filter_key, error in six.iteritems(traced_data):
                error_count["Errors/Count/All"] += error["count"]

                if error["tracker_type"] in error_types:
                    error_count["Errors/Count/AllWeb"] += error["count"]

                    action_key = "Errors/Count/%s" % error_filter_key.split("_|")[0]
                    if action_key not in error_count:
                        error_count[action_key] = error["count"]
                    else:
                        error_count[action_key] += error["count"]

                    if error["tracker_type"] == external_error:
                        action_key = "Errors/Type:%s/%s" % (error["status_code"], error_filter_key.split("_|")[0])
                        if action_key not in error_count:
                            error_count[action_key] = error["count"]
                        else:
                            error_count[action_key] += error["count"]
                else:
                    error_count["Errors/Count/AllBackground"] += 1

        parse_error_trace(self.__traced_errors)
        parse_error_trace(self.__traced_external_errors)

        stat_value = []
        for key, value in six.iteritems(error_count):
            upload_key = {"name": key}
            upload_key_str = '%s' % key
            upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]
            stat_value.append([upload_key, [value]])

        return stat_value

    # stat for error trace data
    # rely to the basic error trace data structure
    def error_trace_data(self):
        """
        :return:
        """
        return [error["item"] for error in six.itervalues(self.__traced_errors)]

    def external_error_data(self):
        return [error["item"] for error in six.itervalues(self.__traced_external_errors)]

    def general_trace_metric(self, metric_name_ids):
        """
        :return:
        """
        result = []
        for key, value in six.iteritems(self.__general_packets):
            extend_keys = key[0].split("|")
            if len(extend_keys) == 1:
                upload_key = {"name": key[0]}
                upload_key_str = '%s' % key[0]
                upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]
                result.append([upload_key, value])
            elif len(extend_keys) == 3:
                # do not replace the metric with id.
                upload_key = {"name": extend_keys[0], "calleeId": extend_keys[1], "calleeName": extend_keys[2]}
                result.append([upload_key, value])

        return result

    def action_trace_data(self):
        """
        :return:
        """
        if not self.__slow_action:
            return []

        trace_data = []

        for traces in six.itervalues(self.__slow_action):
            for trace in traces:
                trace_data.append(trace[0])

        return {"type": "actionTraceData", "actionTraces": trace_data}

    def slow_sql_data(self):
        """
        :return:
        """
        if not self.__slow_sql_packets:
            return []

        result = {"type": "sqlTraceData", "sqlTraces": []}
        maximum = self.__settings.slow_sql_count
        slow_sql_nodes = sorted(six.itervalues(self.__slow_sql_packets), key=lambda x: x.max_call_time)[-maximum:]

        for node in slow_sql_nodes:
            explain_plan = node.slow_sql_node.explain_plan
            params = {"explainPlan": explain_plan if explain_plan else {}, "stacktrace": []}

            if node.slow_sql_node.stack_trace:
                for line in node.slow_sql_node.stack_trace:
                    if len(line) >= 4 and 'tingyun' not in line[0]:
                        params['stacktrace'].append("%s(%s:%s)" % (line[2], line[0], line[1]))

            result['sqlTraces'].append([node.slow_sql_node.start_time, node.slow_sql_node.path,
                                        node.slow_sql_node.metric, node.slow_sql_node.request_uri,
                                        node.slow_sql_node.formatted, node.call_count,
                                        node.total_call_time, node.max_call_time,
                                        node.min_call_time, str(params)])

        return result




"""

"""

import sys
from tingyun.packages import six
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.ammunition.external_tracker import wrap_external_trace, parse_parameters
from tingyun.armoury.ammunition.external_tracker import MALFORMED_URL_ERROR_CODE
from tingyun.armoury.ammunition.external_tracker import OTHER_ERROR_CODE

# python3.4+ has not this module. so do not catch the import error.
from urllib2 import URLError, HTTPError


def wrap_exception(wrapped, url, parameters, *args, **kwargs):
    """
    """

    tracker = current_tracker()
    if tracker is None:
        return wrapped(*args, **kwargs)

    url = url if not callable(url) else url(*args, **kwargs)
    params = parameters if not callable(parameters) else parameters(*args, **kwargs)

    http_status = 500
    error_code = 0
    params = parse_parameters(url, params)

    try:
        rtv = wrapped(*args, **kwargs)
    except URLError:
        error_code = MALFORMED_URL_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params)
        raise
    except HTTPError as err:
        http_status = getattr(err, "code")  # equal to getcode()
        error_code = OTHER_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params)
        raise

    except Exception:
        error_code = OTHER_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params)
        raise

    http_status = rtv.getcode()
    if http_status is not None and int(http_status) != 200:
        tracker.record_external_error(url, error_code, int(http_status), sys.exc_info(), params)

    return rtv


def detect(module):
    def url_opener_open(instance, fullurl, *args, **kwargs):
        """
        :param instance:
        :param fullurl:
        :param args:
        :param kwargs:
        :return:
        """
        if isinstance(fullurl, six.string_types):
            return fullurl
        else:
            return fullurl.get_full_url()

    wrap_external_trace(module, 'OpenerDirector.open', 'urllib2', url_opener_open, exception_wrapper=wrap_exception)



from collections import namedtuple


_ErrorNode = namedtuple('_ErrorNode', ['error_time', 'http_status', "error_class_name", 'uri', 'thread_name',
                                       "message", 'stack_trace', 'request_params', "tracker_type", "referer"])

_ExternalNode = namedtuple('_ExternalNode', ['error_time', 'status_code', 'url', 'thread_name', 'tracker_type',
                                             'error_class_name', 'stack_trace', 'request_params',
                                             'http_status'])


class ErrorNode(_ErrorNode):
    """

    """
    pass


class ExternalErrorNode(_ExternalNode):
    """
    """
    pass




""" define some logistics wrapper

"""

import logging
import sys

from tingyun.packages.wrapt.wrappers import (FunctionWrapper as _FunctionWrapper)
from tingyun.packages.wrapt.wrappers import BoundFunctionWrapper as _BoundFunctionWrapper

from tingyun.packages.wrapt.wrappers import wrap_object as _wrap_object, wrap_function_wrapper as _wrap_function_wrapper
from tingyun.packages.wrapt.wrappers import function_wrapper as _function_wrapper

console = logging.getLogger(__name__)


# ----------------------------------------------------------------------------
# following method are used the wrapt module instead.
# all methods would send from this module, not use the wrapt package directly

wrap_object = _wrap_object

# for wrap the function(bound/unbound)
FunctionWrapper = _FunctionWrapper
BoundFunctionWrapper = _BoundFunctionWrapper
wrap_function_wrapper = _wrap_function_wrapper
function_wrapper = _function_wrapper


def import_module(name):
    """Import module, returning the module after the last dot."""
    __import__(name)
    return sys.modules[name]


def in_function(function):
    @function_wrapper
    def _wrapper(wrapped, instance, args, kwargs):
        if instance is not None:
            args, kwargs = function(instance, *args, **kwargs)
            return wrapped(*args[1:], **kwargs)

        args, kwargs = function(*args, **kwargs)
        return wrapped(*args, **kwargs)

    return _wrapper


def in_function_wrapper(wrapped, function):
    return in_function(function)(wrapped)


def trace_in_function(module, object_path, function):
    return wrap_object(module, object_path, in_function_wrapper, (function,))


def wrap_out_function(function):
    @function_wrapper
    def _wrapper(wrapped, instance, args, kwargs):
        return function(wrapped(*args, **kwargs))
    return _wrapper


def out_function_wrapper(wrapped, function):
    return wrap_out_function(function)(wrapped)


def trace_out_function(module, object_path, function):
    return wrap_object(module, object_path, out_function_wrapper, (function, ))




"""define module for tornado wsgi entrance. but not support tornado core functions

"""

from tingyun.armoury.trigger.wsgi_entrance import wsgi_application_wrapper


def detect_wsgi_entrance(module):
    """
    :param module:
    :return:
    """
    try:
        import tornado
        version = tornado.version
    except Exception:
        version = "xx"

    wsgi_application_wrapper(module.WSGIAdapter, '__call__', ('Tornado', version))




"""this module is implement some wrapper for trace the tornado httpserver module

"""

import sys
import logging
import traceback

from tingyun.battlefield.tracer import Tracer
from tingyun.battlefield.proxy import proxy_instance
from tingyun.logistics.object_name import callable_name
from tingyun.logistics.basic_wrapper import FunctionWrapper

from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.ammunition.function_tracker import FunctionTracker


console = logging.getLogger(__name__)


def request_environ(environ, request):
    """the environ in the default environ may not include our need environment variable when apply to wsgi application
    :param environ: returned environment from the container request.
    :param request:
    :return:
    """
    # we need this the build the request information. but not included in the environment.
    ret = dict(environ)
    ret["REQUEST_URI"] = request.uri
    ret["HTTP_REFERER"] = request.headers.get('Referer', '')
    ret['HTTP_X_QUEUE_START'] = request.headers.get("X-Queue-Start")

    if not environ:
        ret['PATH_INFO'] = request.path
        ret['SCRIPT_NAME'] = ""
        ret['QUERY_STRING'] = request.query

    return ret


def setup_func_for_async_trace(request, name, group='async.wait'):
    """used to suspend function between async call with function
    thread_mode: this indicate, relate the tracer to the thread.this should be used with resume_async_trace func

    request: the HTTPRequest for connection
    """
    tracer = getattr(request, '_self_tracer', None)
    if not tracer:
        console.warning("tracker lost when trace the request call chain with async, this trace for function time"
                        " metric will be interrupted. %s", ''.join(traceback.format_stack()[:-1]))
        return

    if getattr(request, "_self_async_function_tracker", None):
        console.warning("The last tracker for time metric not finished, but this should not happened."
                        "this maybe some logical error in agent. %s", ''.join(traceback.format_stack()[:-1]))
        return

    # console.info("set func for async trace with name %s", name)
    request._self_async_function_tracker = FunctionTracker(tracer, name, group)
    request._self_async_function_tracker.__enter__()


def finish_async_func_trace(request, group=None):
    """calculate the time metric for async call between func
    thread_mode: this indicate, relate the tracer to the thread.this should be used with setup_async_trace func

    request: the HTTPRequest for connection
    """
    tracer = getattr(request, '_self_tracer', None)
    if not tracer:
        console.warning("tracker lost when trace the request call chain with async, this trace for function time"
                        " metric will be interrupted. %s", ''.join(traceback.format_stack()[:-1]))
        return

    try:
        # console.info("finish async func trace with group %s", group)
        if request._self_async_function_tracker:
            request._self_async_function_tracker.__exit__(None, None, None)
            tracer.async_func_trace_time += request._self_async_function_tracker.duration

    finally:
        request._self_async_function_tracker = None

    return tracer


def stop_request_tracer(request, exc_type=None, exc_val=None, exc_tb=None, segment=None):
    """
    """
    tracer = getattr(request, '_self_tracer', None)

    try:
        if not tracer:
            console.debug("tracker lost when trace the request call chain with async, this trace for function time"
                          " metric will be interrupted. %s", ''.join(traceback.format_stack()[:-1]))
            return

        # deal the case which the tracer hold by current request saved in thread cache
        thread_tracer = current_tracker()
        if tracer == thread_tracer:
            tracer.finish_work(exc_type, exc_val, exc_tb, False)
        else:
            tracer.finish_work(exc_type, exc_val, exc_tb, True)
    except Exception as err:
        console.exception("Tornado raise error when stop the trace. %s, %s", segment, err)
    finally:
        request._self_request_finished = True
        request._self_tracer = None
        request._self_async_function_tracker = None

    # console.info("stop request tracer with segment %s", segment)


def generate_tracer(request, framework='Tornado'):
    """
    :param request: the http request for client
    :return:
    """
    tracer = Tracer(proxy_instance(), request_environ({}, request), framework)
    if not tracer.enabled:
        # console.debug("Agent not prepared, skip trace this request now.")
        return None

    try:
        tracer.start_work()
        tracer.drop_tracker()  # drop it from thread cache

        request._self_request_finished = False
        request._self_tracer = tracer
        request._self_async_function_tracker = None
    except Exception as err:
        stop_request_tracer(request, *(sys.exc_info() + ("generate-tracker-exception", )))
        console.exception("Error occurred, when generate a tracker for tornado. %s", err)
        raise

    return tracer


def connection_on_headers_wrapper(wrapped):
    """if client request server without body, so the request in connection object maybe None at there
    """
    def wrapper(wrapped, adapter, args, kwargs):
        """
        :param wrapped: the wrapped function `HTTPConnection._on_headers`
        :param adapter: the instance of the `HTTPConnection`
        :param args: args for `HTTPConnection._on_headers`
        :param kwargs: kwargs for `HTTPConnection._on_headers`
        :return: return of `HTTPConnection._on_headers` method
        """
        # add some ensurance for potential error, if this occurred, that indicate some wrong with last trace.
        # Then we drop the last trace and ignore this trace now.
        tracer = current_tracker()
        if tracer:
            console.warning("Unexpected situation arise, but no side effect to use the agent. That's only indicate "
                            "some illogicality tracker in tracer. if this continue, please report to us, thank u.")
            tracer.drop_tracker()
            return wrapped(*args, **kwargs)

        ret = wrapped(*args, **kwargs)

        # for version3.x.x
        if (hasattr(adapter, "stream") and adapter.stream.closed()) or \
           (hasattr(adapter, "_request_finished") and adapter._request_finished):

            # console.debug("stream closed(%s). in version 3", adapter.stream.closed())
            # console.debug("request finished. %s in version 3", adapter._request_finished)
            return ret

        # for version4.x.x
        if hasattr(adapter, "connection") and adapter.connection.stream.closed():
            # console.debug("stream closed(%s). .in version 4", adapter.stream.closed())
            return ret

        # Because of the asynchronous and single thread feature. we store the tracer in the request.
        # we use _self_ to prevent conflict to the wrapped func namespace.
        # request maybe empty.
        if hasattr(adapter, "connection"):
            request = adapter.request
        else:
            request = adapter._request

        if not request or hasattr(request, '_self_tracer'):
            # console.debug("request(%s) or has _self_tracer(%s)", request, hasattr(request, '_self_tracer'))
            return ret

        # when request with no body, the wrapped function will call Application.__call__, so the tracer maybe create
        # then we should skip.
        if hasattr(request, '_self_tracer'):
            # console.info("on-header returned with request has tracer before")
            return ret

        tracer = generate_tracer(request)
        if not tracer or not tracer.enabled:
            return ret

        # iostream can not fetch the _request object, so we just set a callback to the stream associate with the request
        # with closure
        def _stream_close_callback():
            if finish_async_func_trace(request, ""):
                return

            stop_request_tracer(request, segment='stream-close-callback')

        # version 4
        if hasattr(adapter, "connection"):
            adapter.connection.stream._self_stream_close_callback = _stream_close_callback
        else:
            adapter.stream._self_stream_close_callback = _stream_close_callback

        tracer.set_tracker_name(callable_name(wrapped))
        # at the start beginning, we set the track to the request object for next call-step use
        setup_func_for_async_trace(request, "connection.header-to-body")
        return ret

    return FunctionWrapper(wrapped, wrapper)


def connect_on_request_body_wrapper(wrapped):
    """
    """
    def wrapper(wrapped, adapter, args, kwargs):
        """
        """
        # in tornado 4.x.x
        if hasattr(adapter, "connection"):
            request = adapter.request
        else:
            request = adapter._request

        tracer = finish_async_func_trace(request, group='request-body-in')
        if not tracer:
            # console.info("body do not get the tracer...")
            return wrapped(*args, **kwargs)

        try:
            # console.info("execute request body...")
            ret = wrapped(*args, **kwargs)
        except Exception as err:
            console.exception("Tornado raise error in HTTPConnection on_request_body. %s", err)
            stop_request_tracer(request, *(sys.exc_info() + ("request-body-exception", )))
            raise
        else:
            if request._self_request_finished:
                return ret

            if not request.connection.stream.writing():
                stop_request_tracer(request, segment='stream-not-writing')
                return ret

            setup_func_for_async_trace(request, name="connection.body-to-finish")

        return ret

    return FunctionWrapper(wrapped, wrapper)


def connection_finish_request_wrapper(wrapped):
    """the wrapped function maybe called more than once.
    """
    def wrapper(wrapped, adapter, args, kwargs):
        """
        """
        # in tornado 4.x.x
        if hasattr(adapter, "connection"):
            request = adapter.request
        else:
            request = adapter._request

        tracer = getattr(request, "_self_tracer", None)

        if not tracer:
            # console.debug("tracer is not found maybe finished before.")
            return wrapped(*args, **kwargs)

        if request._self_request_finished:
            # console.debug("tracer for current request is finished")
            return wrapped(*args, **kwargs)

        tracer = finish_async_func_trace(request, "request-finish-in")
        if not tracer:
            # console.debug("resumed a none tracer....")
            return wrapped(*args, **kwargs)

        try:
            # console.info("execute connect finish finished....")
            ret = wrapped(*args, **kwargs)
        except Exception as _:
            stop_request_tracer(request, *(sys.exc_info() + ("connect-finished-with-exceptiono", )))
            raise
        else:
            # in tornado 4.x.x, the application.__call__ will not be called directly. but call the `execute` to call
            # the RequestHandler. so we should finish our tracker at there.
            if hasattr(adapter, "connection"):
                stop_request_tracer(request, segment='finish-app-call')

        return ret

    return FunctionWrapper(wrapped, wrapper)


def iostream_close_callback_wrapper(wrapped):
    """if client communicate with server user keep-alive protocol, the callback can not be actived.
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        """
        if not instance.closed():
            # console.debug("stream not closed, skip to finish the trace.")
            return wrapped(*args, **kwargs)

        if instance._pending_callbacks != 0:
            # console.debug("pending callback, skip to finish the trace.")
            return wrapped(*args, **kwargs)

        callback = getattr(instance, '_self_stream_close_callback', None)
        instance._self_stream_close_callback = None

        if callback:
            callback()

        return wrapped(*args, **kwargs)

    return FunctionWrapper(wrapped, wrapper)




"""this module is used to process the cross application/transaction trace with processing the header.
"""

import logging


console = logging.getLogger(__name__)


def process_header(tracker, headers):
    """
    :param tracker:
    :param headers:
    :return:
    """
    action_trace_invoke = 0
    if not tracker.enabled:
        return

    if not tracker.call_tingyun_id:
        return

    tingyun_ids = tracker._tingyun_id.split("|")
    if tingyun_ids < 2:
        console.debug("tingyun id is not satisfied, if this continue please contact us.")
        return

    # 1 current trace overflow the action threshold
    # 2 the called service has the overflow the action threshold.
    if (tracker.duration >= tracker.settings.action_tracer.action_threshold) or \
       (tracker._called_traced_data and hasattr(tracker._called_traced_data, "tr")):
        action_trace_invoke = 1

    services = {'ex': tracker.external_time, 'rds': tracker.redis_time, 'mc': tracker.memcache_time,
                'mon': tracker.mongo_time, 'db': tracker.db_time}

    duration = tracker.duration
    code_time = duration - sum([t for t in services.values() if t >= 0])
    trace_data = {
        "id": tingyun_ids[1].encode("utf8"),
        "action": tracker.path.encode("utf8"),
        "trId": tracker.generate_trace_guid().encode("utf8"),
        "time":
            {
            'duration': duration,
            'qu': tracker.queque_time,
            'code': code_time,
        }
    }

    for key, value in services.iteritems():
        if value >= 0:
            trace_data["time"][key] = value

    if 0 != action_trace_invoke:
        trace_data["tr"] = action_trace_invoke

    if tracker.call_req_id:
        trace_data["r"] = tracker.call_req_id.encode("utf8")

    headers.append(("X-Tingyun-Tx-Data", "%s" % trace_data))




"""define a module for detect framework web2py. only support 2.8.1 upon
"""

from tingyun.armoury.trigger.wsgi_entrance import wsgi_application_wrapper
from tingyun.armoury.ammunition.web2py_trakcer import trace_serve_controller
from tingyun.armoury.ammunition.function_tracker import wrap_function_trace


def detect_wsgi_entrance(module):
    """
    :param module:
    :return:
    """
    version = 'xx'

    try:
        import gluon.main
        version = gluon.main.web2py_version.strip()
        version = version[: 5]
    except Exception as _:
        pass

    wsgi_application_wrapper(module, 'wsgibase', ('web2py', version))
    trace_serve_controller(module, 'serve_controller')


def detect_compileapp(module):
    """
    :param module:
    :return:
    """
    def get_name_in_controller(controller, function, environment):
        """more detail in gluon.compileapp.run_controller_in
        """
        return '%s.%s' % (controller, function)

    def get_name_in_model(environment):
        """more detail in gluon.compileapp.run_model_in
        """
        return '%s.%s' % (environment['request'].controller, environment['request'].function)

    def get_name_in_view(environment):
        """more detail in gluon.compileapp.run-view-in
        """
        return '%s.%s' % (environment['request'].controller, environment['request'].function)

    wrap_function_trace(module, 'run_controller_in', get_name_in_controller, group='web2py.controllers')
    wrap_function_trace(module, 'run_models_in', get_name_in_model, group='web2py.models')
    wrap_function_trace(module, 'run_view_in', get_name_in_view, group='web2py.views')


def detect_template(module):
    """
    :param module:
    :return:
    """
    def get_template_name(filename, *args, **kwargs):
        """
        """
        return filename

    wrap_function_trace(module, 'parse_template', get_template_name, group='template.compile')




"""this module is defined to wrap the thrift.
"""

import sys
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.ammunition.external_tracker import wrap_external_trace, parse_parameters
from tingyun.armoury.ammunition.external_tracker import ILLEGAL_RESPONSE_ERROR_CODE, UNKNOWN_HOST_ERROR_CODE
from tingyun.armoury.ammunition.external_tracker import OTHER_ERROR_CODE


try:
    from thrift.transport.TTransport import TTransportException
except ImportError:
    class TTransportException(Exception):
        pass

try:
    from thrift.protocol.TProtocol import TProtocolException
except ImportError:
    class TProtocolException(Exception):
        pass


def wrap_exception(wrapped, url, parameters, *args, **kwargs):
    """
    """

    tracker = current_tracker()
    if tracker is None:
        return wrapped(*args, **kwargs)

    url = url if not callable(url) else url(*args, **kwargs)
    params = parameters if not callable(parameters) else parameters(*args, **kwargs)

    http_status = 500
    error_code = 0
    params = parse_parameters(url, params)

    try:
        return wrapped(*args, **kwargs)
    except TTransportException:
        error_code = UNKNOWN_HOST_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params)
        raise
    except TProtocolException:
        error_code = ILLEGAL_RESPONSE_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params)
        raise
    except Exception:
        error_code = OTHER_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params)
        raise


def detect_tsocket(module):
    """
    :param module:
    :return:
    """
    def tsocket_url(tsocket, *args, **kwargs):
        # _unix_socket is passed in when TSocket initialized
        # when wrap the specified func, the tsocket is instance.
        url = 'thrift://%s:%s' % (tsocket.host, tsocket.port)

        if tsocket._unix_socket is None:
            url = 'thrift://%s:%s' % (tsocket.host, tsocket.port)

        return url

    wrap_external_trace(module, 'TSocket.open', 'thrift', tsocket_url, exception_wrapper=wrap_exception)


def detect_tsslsocket(module):
    """
    :param module:
    :return:
    """
    def tsocket_ssl_url(tsocket, *args, **kwargs):
        # _unix_socket is passed in when TSSSocket initialized
        # when wrap the specified func, the TSSSocket is instance.
        url = 'thrift//%s:%s' % (tsocket.host, tsocket.port)

        if tsocket._unix_socket is None:
            url = 'thrift//%s:%s' % (tsocket.host, tsocket.port)

        return url

    wrap_external_trace(module, 'TSSLSocket.open', 'thrift', tsocket_ssl_url, exception_wrapper=wrap_exception)



"""This module implements functions for deriving the full name of an object.

"""

import sys
import types
import inspect
import logging
from tingyun.packages import six

console = logging.getLogger(__name__)
cached_module_name = {}


def get_module_name(target):
    module_name = None

    if hasattr(target, '__objclass__'):
        module_name = getattr(target.__objclass__, '__module__', None)

    if module_name is None:
        module_name = getattr(target, '__module__', None)

    # An exception to that is builtins or any types which are implemented in C code. For that we need to grab the module
    # name from the __class__.
    if module_name is None and hasattr(target, '__class__'):
        module_name = getattr(target.__class__, '__module__', None)

    # Finally, if the module name isn't in sys.modules, we will format it to indicate that it is a generated
    # class of some sort where a fake namespace was used. This happens for example with namedtuple classes in Python 3.
    if module_name and module_name not in sys.modules:
        module_name = '<%s>' % module_name

    if not module_name:
        module_name = '<unknown>'

    return module_name


def _object_context_py2(obj):

    cname = None
    fname = None

    if inspect.isclass(obj) or isinstance(obj, type):
        # Old and new style class types.
        cname = obj.__name__

    elif inspect.isfunction(obj):
        # Normal functions and static methods. For a static we method don't know of any way of being able to work out
        # the name of the class the static method is against.
        fname = obj.__name__

    elif inspect.ismethod(obj):
        if obj.im_self is not None:
            cname = getattr(obj.im_self, '__name__', None)
            if cname is None:
                cname = getattr(obj.im_self.__class__, '__name__')

        else:
            cname = obj.im_class.__name__

        fname = obj.__name__

    elif inspect.isbuiltin(obj):
        # Builtin function. Can also be be bound to class to create a method. Uses __self__ instead of im_self. The
        # rules around whether __self__ is an instance or a class type are strange so need to cope with both.

        if obj.__self__ is not None:
            cname = getattr(obj.__self__, '__name__', None)
            if cname is None:
                cname = getattr(obj.__self__.__class__, '__name__')

        fname = obj.__name__

    elif isinstance(obj, types.InstanceType):
        fname = getattr(obj, '__name__', None)
        if fname is None:
            cname = obj.__class__.__name__

    elif hasattr(obj, '__class__'):

        fname = getattr(obj, '__name__', None)
        if fname is not None:
            if hasattr(obj, '__objclass__'):
                cname = obj.__objclass__.__name__
            elif not hasattr(obj, '__get__'):
                cname = obj.__class__.__name__
        else:
            cname = obj.__class__.__name__

    path = ''
    if cname:
        path = cname

    if fname:
        if path:
            path += '.'
        path += fname

    module_name = get_module_name(obj)

    return module_name, path


def _object_context_py3(obj):
    # functions and methods the __qualname__ attribute gives the name.
    path = getattr(obj, '__qualname__', None)

    # If there is no __qualname__ it maybe mean it is a type object of some sort.
    if path is None and hasattr(obj, '__class__'):
        path = getattr(obj.__class__, '__qualname__')

    module_name = get_module_name(obj)

    return module_name, path


def object_context(target):
    """Returns a tuple identifying the supplied object. This will be of
    the form (module, object_path).

    """
    cache_key = str(target)
    if cache_key in cached_module_name:
        return cached_module_name[cache_key]

    if six.PY3:
        details = _object_context_py3(target)
    else:
        details = _object_context_py2(target)

    if str(target) not in cached_module_name:
        cached_module_name[str(target)] = details

    return details


def callable_name(obj, separator=':'):
    """Returns a string name identifying the supplied object. the form is 'module:object_path'.

    If object were a function, then the name would be 'module:function.
    If a class, 'module:class'.
    If a member function, 'module:class.function'.
    """
    context = object_context(obj)

    return separator.join(context)




"""define the interface for browser monitor.

"""

import re
import logging
import random


console = logging.getLogger(__name__)

head_re = re.compile(b'<head[^>]*>', re.IGNORECASE)
body_re = re.compile(b'<body[^>]*>', re.IGNORECASE)
xua_meta_re = re.compile(b"""<\s*meta[^>]+http-equiv\s*=\s*['"]\s*x-ua-compatible\s*['"][^>]*>""", re.IGNORECASE)
charset_meta_re = re.compile(b"""<\s*meta[^>]+charset\s*=\s*['"][^>]*['"][^>]*>""", re.IGNORECASE)
attachment_re = re.compile(b"""<\s*meta[^>]+http-equiv\s*=\s*['"]"""
                           b"""\s*content-disposition\s*['"][^>]*content\s*=\s*(?P<quote>['"])"""
                           b"""\s*attachment(\s*;[^>]*)?(?P=quote)[^>]*>""", re.IGNORECASE)


class TingYunWSGIBrowserRumMiddleware(object):
    """More information come with pep333: https://www.python.org/dev/peps/pep-0333
    """
    def __init__(self, tracker, application, start_response, environ):
        self.tracker = tracker
        self.ignore_rum = True
        self.application = application
        self.environ = environ
        self.org_start_response = start_response
        self.sample_ratio = 0.0
        self.enable = self.is_enable_auto_rum()
        self.allowed_content = ['text/html']

        self.status = None
        self.response_headers = []
        self.content_length = None
        self.args = ()
        self.writer = None

    def is_enable_auto_rum(self):
        if not self.tracker.enabled or not self.tracker.settings:
            return False

        if not self.tracker.settings.rum.enabled:
            return False

        self.sample_ratio = self.tracker.settings.rum.sample_ratio
        if self.sample_ratio < random.random():
            return False

        return True

    def write_headers(self):
        """
        :return:
        """
        header = []
        if self.content_length is not None:
            for key, value in self.response_headers:
                if key != 'Content-Length':
                    header.append((key, value))
                else:
                    # data length changed by agent.
                    header.append((key, "%s" % self.content_length))

            self.response_headers = header

        self.writer = self.org_start_response(self.status, self.response_headers, *self.args)

    def write(self, data):
        """define this for delay call real start_response func.
            WSGI specification define the func just for compatible with old version framework.

            This func should not be used, but some old framework may use it. so avoid some issue, skip auto insert js.
        :return:
        """
        self.ignore_rum = True

        if not self.writer:
            self.write_headers()

        console.debug("framework use the write method..")
        return self.writer(data)

    def start_response(self, status, response_headers, *args):
        """Note: The application may call start_response more than once,
                 if and only if the exc_info argument is provided.
            It is a fatal error to call start_response without the exc_info argument
            if start_response has already been called within the current invocation of the application
        :return: write func which returned by wsgi start_response
        """
        self.status = status
        self.response_headers = response_headers
        self.args = args
        enabled = self.is_enable_auto_rum()
        header_dict = dict([(str(key).lower(), value) for (key, value) in response_headers])

        if not enabled:
            self.write_headers()
            self.ignore_rum = True
            return self.write

        def is_insertable():
            """follow rules will return false
               . content specified encoding, wsgi application should not return encoded response
                    and maybe has some performance issue
               . attachment for file download
               . content-type is not html response
               . invalid content-length
            """
            if 'content-length' in header_dict:
                try:
                    int(header_dict['content-length'])
                except ValueError:
                    return False

            content_type = header_dict.get("content-type", None)
            if not content_type or content_type.split(';')[0] not in self.allowed_content:
                return False

            content_disposition = header_dict.get('content-disposition', None)
            if content_disposition is not None and content_disposition.split(';')[0].strip().lower() == 'attachment':
                return False

            return True

        if is_insertable():
            self.ignore_rum = False
            self.content_length = header_dict.get('content-length', None)

        if self.ignore_rum:
            self.ignore_rum = True
            self.write_headers()
            return self.write

        # at last. we return the Compatible write just for delay call start_response for avoiding the gateway/server
        # flush the header when first iterable data prepared.
        # and we will change the data length if it exist.
        return self.write

    def do_auto_monitor(self, data):
        """
        :param data: response data.
        :return:
        """
        buffer_data = self.insert_js_to_html(data, self.tracker.fork_browser_js_head())
        insert_data_length = len(buffer_data) - len(data)

        if self.content_length and insert_data_length:
            self.content_length = int(self.content_length)
            self.content_length += insert_data_length

        # we just check for the first data block with 64kb
        self.ignore_rum = True
        return insert_data_length, buffer_data

    def __call__(self, *args, **kwargs):
        """
        :param args:
        :param kwargs:
        :return:
        """
        for ret in self.application(self.environ, self.start_response):
            # maybe some blocks are not header parts.
            if self.ignore_rum:
                yield ret
                continue

            # WSGI specification note: start_response should not be called before get first block of data.
            # just do insert the js when headers are not send because we need to change the length if necessary.
            if not self.writer:
                if not ret:
                    continue

                _, buffer_data = self.do_auto_monitor(ret)
                self.write_headers()
                self.ignore_rum = True
                yield buffer_data
            else:
                console.debug("Warning! Ignore the rum because of headers are send before.")
                yield ret

        if not self.writer:
            self.write_headers()
            self.ignore_rum = True

    def insert_js_to_html(self, data, agent_js, search_size=65536):
        """
        :param data:
        :param js_seg:
        :param search_size:32Kb
        :return:
        """
        body = body_re.search(data[:search_size])
        if not body:
            return data

        # skip the attachment request part
        if attachment_re.search(data[:search_size]):
            return data

        # we should insert after page charset and the compatible mode
        xua = xua_meta_re.search(data[:search_size])
        charset = charset_meta_re.search(data[:search_size])
        position = max(xua and xua.end() or 0, charset and charset.end() or 0)

        def combine_data(position):
            return b''.join((data[:position], agent_js, data[position:]))

        if position:
            return combine_data(position)

        head_meta = head_re.search(data[:search_size])
        if head_meta:
            return combine_data(head_meta.end())
        else:
            # head not define in pages
            return combine_data(body.start())




"""Define this module for basic armory for bottle

"""

import logging
from tingyun.armoury.trigger.wsgi_entrance import wsgi_application_wrapper
from tingyun.armoury.ammunition.function_tracker import wrap_function_trace, FunctionTracker
from tingyun.logistics.basic_wrapper import FunctionWrapper, trace_out_function
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.logistics.object_name import callable_name

console = logging.getLogger(__name__)


def detect_wsgi_entrance(module):
    """
    :param module:
    :return:
    """
    import bottle
    version = getattr(bottle, "__version__", 'xx')

    wsgi_application_wrapper(module.Bottle, '__call__', ('bottle', version))


def detect_templates(module):
    """
    :param module:
    :return:
    """
    # detect the template render
    if hasattr(module, 'SimpleTemplate'):
        wrap_function_trace(module, 'SimpleTemplate.render')

    if hasattr(module, 'MakoTemplate'):
        wrap_function_trace(module, 'MakoTemplate.render')

    if hasattr(module, 'CheetahTemplate'):
        wrap_function_trace(module, 'CheetahTemplate.render')

    if hasattr(module, 'Jinja2Template'):
        wrap_function_trace(module, 'Jinja2Template.render')

    if hasattr(module, 'SimpleTALTemplate'):
        wrap_function_trace(module, 'SimpleTALTemplate.render')


def route_callback_wrapper(wrapped):
    """wrap the route callback for trace the `route` performance
    :return:
    """
    def wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()
        if not tracker:
            return wrapped(*args, **kwargs)

        tracker.set_tracker_name(callable_name(wrapped), priority=3)
        with FunctionTracker(tracker, callable_name(wrapped)):
            try:
                return wrapped(*args, **kwargs)
            except Exception as _:
                tracker.record_exception()
                raise

    return FunctionWrapper(wrapped, wrapper)


def detect_app_components(module):
    """
    :param module:
    :return:
    """
    if hasattr(module, 'Route') and hasattr(module.Route, '_make_callback'):
        trace_out_function(module, 'Route._make_callback', route_callback_wrapper)



"""this module implement the metric attribution definition

"""
from collections import namedtuple

# for metric
Metric = namedtuple('Metric', ['name', 'scope'])
ApdexMetric = namedtuple('ApdexMetric', ['name', 'satisfying', 'tolerating', 'frustrating', 'apdex_t'])
TimeMetric = namedtuple('TimeMetric', ['name', 'scope', 'duration', 'exclusive'])

# for error trace
TracedError = namedtuple('TracedError', ['error_filter_key', 'tracker_type', "trace_data"])
TracedExternalError = namedtuple("TracedExternalError", ['error_filter_key', "trace_data", 'tracker_type',
                                                         "status_code"])


RootNode = namedtuple('RootNode', ['start_time', 'request_uri', 'duration', 'name', 'trace_data'])
TraceNode = namedtuple('TraceNode', ['start_time', 'end_time', 'metric_name', 'call_url', 'call_count', 'class_name',
                                     'method_name', 'params', 'children'])


def node_start_time(root, node):
    return int((node.start_time - root.start_time) * 1000.0)


def node_end_time(root, node):
    return int((node.end_time - root.start_time) * 1000.0)


def node_duration_time(node):
    return int((node.end_time - node.start_time) * 1000.0)



from __future__ import print_function

import os
from tingyun.logistics.exceptions import CommandlineParametersException


class Command(object):
    """
    """
    def __init__(self):
        """
        """
        self.name = "generate-config"  # order format with commandline
        self.options = '[license_key] filename'
        self.description = "Generate the config file with license(optional) with filename"

    def execute(self, args):
        """generate default configuration to specified path.
        :param args:
        :return:
        """
        if len(args) < 1:
            raise CommandlineParametersException()

        from tingyun import __file__ as package_root

        package_root = os.path.dirname(package_root)
        config_file = os.path.join(package_root, 'tingyun.ini')
        default_config = open(config_file, 'r')
        content = default_config.read()

        if 2 == len(args):
            output_file = open(args[1], "w")
        else:
            output_file = open(args[0], "w")

        if 2 == len(args):
            content = content.replace('** YOUR-LICENSE-KEY **', args[0])
            print("""
                        ================ Messages ==============
                  You use license key: %s, to generate config file: %s
                  """ % (args[0], args[1]))
        elif 1 == len(args):
            print("""
                        ================ Messages ==============
                  You use license key: , to generate config file: %s.
                  Before you use the python agent, you should type the license key into config file
                  """ % args[0])

        output_file.write(content)
        output_file.close()
        default_config.close()




"""define detector for detect stack context module
"""

import logging
from tingyun.logistics.object_name import callable_name
from tingyun.logistics.basic_wrapper import wrap_function_wrapper
from tingyun.armoury.ammunition.tornado_4.utils import create_tracker_aware_fxn, obtain_current_tracker, record_exception

console = logging.getLogger(__name__)
fxn_cache = []


def trace_wrap(wrapped, instance, args, kwargs):
    """
    """
    def _fxn_arg_extractor(fn, *args, **kwargs):
        return fn

    unwrapped_fxn = _fxn_arg_extractor(*args, **kwargs)
    wrapped_fxn = wrapped(*args, **kwargs)
    should_trace = callable_name(unwrapped_fxn) not in fxn_cache
    if should_trace:
        fxn_cache.append(callable_name(unwrapped_fxn))

    tracker_aware_fxn = create_tracker_aware_fxn(wrapped_fxn, fxn_for_name=unwrapped_fxn, should_trace=should_trace)

    if tracker_aware_fxn is None:
        return wrapped_fxn

    tracker_aware_fxn._wrapped = True
    tracker_aware_fxn._self_tracker = obtain_current_tracker()

    return tracker_aware_fxn


def trace_handle_exception(wrapped, instance, args, kwargs):
    """
    """
    def _exc_extractor(tail, exc, *args, **kwargs):
        return exc

    exc = _exc_extractor(*args, **kwargs)

    record_exception(exc)
    return wrapped(*args, **kwargs)


def detect_stack_context(module):
    """
    """
    wrap_function_wrapper(module, 'wrap', trace_wrap)
    wrap_function_wrapper(module, '_handle_exception', trace_handle_exception)



"""this module implement data packets for packager

"""
import operator


class TimePackets(list):
    """six element of list
    """
    def __init__(self, call_count=0, total_call_time=0.0, total_exclusive_call_time=0.0, max_call_time=0.0,
                 min_call_time=0.0, sum_of_squares=0.0):
        super(TimePackets, self).__init__([call_count, total_call_time, total_exclusive_call_time, max_call_time,
                                           min_call_time, sum_of_squares])
    call_count = property(operator.itemgetter(0))
    total_call_time = property(operator.itemgetter(1))
    total_exclusive_call_time = property(operator.itemgetter(2))
    min_call_time = property(operator.itemgetter(4))
    max_call_time = property(operator.itemgetter(3))
    sum_of_squares = property(operator.itemgetter(5))

    def merge_packets(self, other):
        """Merge data from another instance of this object."""

        self[1] += other[1]
        self[2] += other[2]
        self[4] = self[0] and min(self[4], other[4]) or other[4]
        self[3] = max(self[3], other[3])
        self[5] += other[5]

        # Must update the call count last as update of the minimum call time is dependent on initial value.
        self[0] += other[0]

    def merge_original_time_metric(self, duration, exclusive=None):
        """Merge original time value."""

        if exclusive is None:
            exclusive = duration

        self[1] += duration
        self[2] += exclusive
        self[3] = max(self[3], duration)
        self[4] = self[0] and min(self[4], duration) or duration
        # self[3] = self[0] and min(self[3], duration) or duration
        # self[4] = max(self[4], duration)
        self[5] += duration ** 2

        # Must update the call count last as update of the minimum call time is dependent on initial value.
        self[0] += 1

    def merge_time_metric(self, metric):
        """Merge data from a time metric object."""
        self.merge_original_time_metric(int(metric.duration), int(metric.exclusive))


class ApdexPackets(list):
    """
    """
    def __init__(self, satisfying=0, tolerating=0, frustrating=0, apdex_t=0.0):
        super(ApdexPackets, self).__init__([satisfying, tolerating, frustrating, apdex_t])

    satisfying = property(operator.itemgetter(0))
    tolerating = property(operator.itemgetter(1))
    frustrating = property(operator.itemgetter(2))

    def merge_packets(self, other):
        """Merge data from another instance of this object."""

        self[0] += other[0]
        self[1] += other[1]
        self[2] += other[2]

    def merge_apdex_metric(self, metric):
        """Merge data from an apdex metric object."""

        self[0] += metric.satisfying
        self[1] += metric.tolerating
        self[2] += metric.frustrating


class SlowSqlPackets(list):

    def __init__(self):
        super(SlowSqlPackets, self).__init__([0, 0, 0, 0, None])

    call_count = property(operator.itemgetter(0))
    total_call_time = property(operator.itemgetter(1))
    min_call_time = property(operator.itemgetter(2))
    max_call_time = property(operator.itemgetter(3))
    slow_sql_node = property(operator.itemgetter(4))

    def merge_packets(self, other):
        """Merge data from another instance of this object."""

        self[1] += other[1]
        self[2] = self[0] and min(self[2], other[2]) or other[2]
        self[3] = max(self[3], other[3])

        if self[3] == other[3]:
            self[4] = other[4]

        # Must update the call count last as update of the minimum call time is dependent on initial value.
        self[0] += other[0]

    def merge_slow_sql_node(self, node):
        """Merge data from a slow sql node object."""

        duration = node.duration

        self[1] += duration
        self[2] = self[0] and min(self[2], duration) or duration
        self[3] = max(self[3], duration)

        if self[3] == duration:
            self[4] = node

        # Must update the call count last as update of the minimum call time is dependent on initial value.
        self[0] += 1




"""this module is defined to detect the httplib2 module
"""

import sys
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.ammunition.external_tracker import wrap_external_trace, process_cross_trace, parse_parameters
from tingyun.armoury.ammunition.external_tracker import MALFORMED_URL_ERROR_CODE, UNKNOWN_HOST_ERROR_CODE
from tingyun.armoury.ammunition.external_tracker import CONNECT_EXCEPTION_ERROR_CODE, TIMEOUT_EXCEPTION_ERROR_CODE
from tingyun.armoury.ammunition.external_tracker import CLIENT_PROTOCOL_ERROR_CODE, ILLEGAL_RESPONSE_ERROR_CODE
from tingyun.armoury.ammunition.external_tracker import OTHER_ERROR_CODE, SSL_EXCEPTION_ERROR_CODE


try:
    from httplib2 import HttpLib2ErrorWithResponse
except ImportError:
    class HttpLib2ErrorWithResponse(Exception):
        pass

try:
    from httplib2 import RedirectMissingLocation
except ImportError:
    class RedirectMissingLocation(Exception):
        pass

try:
    from httplib2 import FailedToDecompressContent
except ImportError:
    class FailedToDecompressContent(Exception):
        pass

try:
    from httplib2 import RelativeURIError
except ImportError:
    class RelativeURIError(Exception):
        pass

try:
    from httplib2 import ServerNotFoundError
except ImportError:
    class ServerNotFoundError(Exception):
        pass

try:
    from httplib2 import CertificateValidationUnsupported
except ImportError:
    class CertificateValidationUnsupported(Exception):
        pass

try:
    from httplib2 import SSLHandshakeError
except ImportError:
    class SSLHandshakeError(Exception):
        pass

try:
    from httplib2 import CertificateHostnameMismatch
except ImportError:
    class CertificateHostnameMismatch(Exception):
        pass


def wrap_exception(wrapped, url, parameters, *args, **kwargs):
    """
    """

    tracker = current_tracker()
    if tracker is None:
        return wrapped(*args, **kwargs)

    url = url if not callable(url) else url(*args, **kwargs)
    params = parameters if not callable(parameters) else parameters(*args, **kwargs)

    http_status = 500
    error_code = 0
    params = parse_parameters(url, params)
    try:
        resp, content = wrapped(*args, **kwargs)
    except (CertificateHostnameMismatch, SSLHandshakeError, CertificateValidationUnsupported):
        error_code = SSL_EXCEPTION_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params)
        raise
    except ServerNotFoundError:
        error_code = UNKNOWN_HOST_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params)
        raise
    except (RelativeURIError, RedirectMissingLocation):
        error_code = MALFORMED_URL_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params)
        raise
    except (FailedToDecompressContent, HttpLib2ErrorWithResponse):
        error_code = ILLEGAL_RESPONSE_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params)
        raise
    except Exception:
        error_code = OTHER_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params)
        raise

    if int(getattr(resp, 'status', 0)) != 200:
        tracker.record_external_error(url, error_code, getattr(resp, 'status', 500), sys.exc_info(), params)

    return resp, content


def detect_httplib2_http(module):
    """
    :param module:
    :return:
    """
    def http_url(instance, uri, method, *args, **kwargs):
        """
        :param instance:Http() instance
        :param uri:
        :param method:
        :param args:
        :param kwargs:
        :return:
        """
        return uri

    def parse_params(uri, method="GET", body=None, headers=None, **kwargs):
        """
        """
        _args = (uri, )
        _kwarg = kwargs

        if not kwargs:
            _kwarg = {}

        _kwarg["method"] = method
        _kwarg["body"] = body
        _kwarg["headers"] = process_cross_trace(headers)

        return _args, _kwarg

    return wrap_external_trace(module, "Http.request", 'httplib2', http_url, params=parse_params,
                               exception_wrapper=wrap_exception)


def detect_http_connect_with_timeout(module):
    """
    :param module:
    :return:
    """
    def http_with_timeout_url(instance, *args, **kwargs):
        """
        :param instance: HTTPConnectionWithTimeout instance
        :param args:
        :param kwargs:
        :return:
        """
        url = "http://%s" % instance.host
        if instance.port:
            url = 'http://%s:%s' % (instance.host, instance.port)

        return url

    wrap_external_trace(module, "HTTPConnectionWithTimeout.connect", 'httplib2', http_with_timeout_url)


def detect_https_connect_with_timeout(module):
    """
    :param module:
    :return:
    """
    def https_with_timeout_url(instance, *args, **kwargs):
        """
        :param instance: HTTPSConnectionWithTimeout instance
        :param args:
        :param kwargs:
        :return:
        """
        url = "http://%s" % instance.host
        if instance.port:
            url = 'http://%s:%s' % (instance.host, instance.port)

        return url

    wrap_external_trace(module, "HTTPSConnectionWithTimeout.connect", 'httplib2', https_with_timeout_url)



from __future__ import print_function

""" Define army dispatch center for kinds of corps
"""

import sys
import os


class Dispatcher(object):
    """Define the commands dispatcher executor

    """
    def __init__(self, argv=None):
        """
        :return:
        """
        self.argv = argv or sys.argv[:]
        self.commands = {}

    def load_tingyun_entrypoint(self):
        """Load the entrypoint defined in setup
        :return:
        """
        try:
            import pkg_resources
        except ImportError:
            return

        for entrypoint in pkg_resources.iter_entry_points(group='tingyun.commander:launch_commanding_elevation'):
            __import__(entrypoint.module_name)

    def help_text(self, command=None):
        """The top level message from dispatcher
        :param command:  command instance
        :return:
        """
        if command is None:
            print("Support commands are:  ")

            for name, cmd in self.commands.iteritems():
                print("    %s %s" % (cmd.name, cmd.options))
        else:
            print("Command Usage:")
            print("    %s %s " % (command.name, command.options))
            print("       %s" % command.description)

    def _load_commands(self):
        """load the commands from the commands box.
        :return: the mapping of commands names with Command object
        """
        cmd_box = os.path.join(__path__[0], 'commands')
        try:
            commands = [f[:-3] for f in os.listdir(cmd_box) if not f.startswith('_') and f.endswith('.py')]
        except OSError:
            commands = []

        for cmd in commands:
            name = "tingyun.commander.commands.%s" % cmd
            __import__(name)
            self.commands[cmd] = sys.modules[name].Command()

    def execute(self):
        """
        :return:
        """
        self._load_commands()

        second_cmd = 'help'
        if len(self.argv) > 1:
            second_cmd = self.argv[1]

        second_cmd = "_".join(second_cmd.split("-"))
        if second_cmd == 'help' or second_cmd not in self.commands.keys():
            self.help_text()
            return

        try:
            self.commands[second_cmd].execute(self.argv[2:])
        except Exception as err:
            print(err)
            self.help_text(self.commands[second_cmd])


def launch_commanding_elevation():
    """ Define corps dispatch center entrance.
        Dispatch orders
    :return:
    """
    dsp = Dispatcher()
    dsp.load_tingyun_entrypoint()
    dsp.execute()



"""the module wrapt main entrance

"""

import logging

from tingyun.armoury.ammunition.mongo_tracker import wrap_mongo_trace
from tingyun.armoury.ammunition.function_tracker import wrap_function_trace

console = logging.getLogger(__name__)
_methods = ['count', 'create_index', 'distinct', 'drop', 'drop_index', 'drop_indexes', 'ensure_index', 'find',
            'find_and_modify', 'find_one', 'group', 'index_information', 'inline_map_reduce', 'insert', 'map_reduce',
            'options', 'reindex', 'remove', 'rename', 'save', 'update']


def detect_connection(module):
    """
    :param module:
    :return:
    """
    if hasattr(module, "Connection"):
        wrap_function_trace(module, 'Connection.__init__', name='%s:Connection.__init__' % module.__name__)


def detect_mongo_client(module):
    """
    :param module:
    :return:
    """
    if hasattr(module, "MongoClient"):
        wrap_function_trace(module, "MongoClient.__init__", name="%s.MongoClient.__init__" % module.__name__)


def detect_collection(module):
    """
    :param module:
    :return:
    """
    def collection_name(collection, *args, **kwargs):
        """ we get the full name include the database name with dot separated
        :param collection:
        :param args:
        :param kwargs:
        :return:
        """
        return collection.full_name

    if not hasattr(module, "Collection"):
        console.info("module(%s) has not Collection object.", module)
        return

    for m in _methods:
        if not hasattr(module.Collection, m):
            console.info("Collection has not %s method", m)
            continue

        wrap_mongo_trace(module, "Collection.%s" % m, schema=collection_name, method=m)




import logging
import os
import sys
import threading
import weakref
import thread


console = logging.getLogger(__name__)


class Knapsack(object):
    """
    """
    _instance = None
    _instance_lock = threading.Lock()

    def __init__(self):
        self._cache = weakref.WeakValueDictionary()
        console.debug("init tracker with thread %s", thread.get_ident())

    @staticmethod
    def singleton_instance():
        """
        :return: singleton instance
        """
        if Knapsack._instance:
            return Knapsack._instance

        instance = None
        with Knapsack._instance_lock:
            if not Knapsack._instance:
                instance = Knapsack()
                Knapsack._instance = instance

                console.info('Creating instance of tracker Knapsack in  process %d.', os.getpid())

        return instance

    def current_thread_id(self):
        """get the current thread id
        :return: thread id
        """
        greenlet = sys.modules.get('greenlet')

        if greenlet:
            # Compatible greenlet frame
            current = greenlet.getcurrent()
            if current is not None and current.parent:
                return id(current)

        return thread.get_ident()

    def save_tracker(self, tracker):
        """
        :param tracker:
        :return:
        """
        if tracker.thread_id in self._cache:
            raise RuntimeError("tracker already exist.")

        self._cache[tracker.thread_id] = tracker

    def get_active_threads(self):
        """
        """
        for thread_id, frame in sys._current_frames().items():
            tracker = self._cache.get(thread_id)
            if tracker is not None:
                if tracker.background_task:
                    yield tracker, thread_id, 'Background', frame
                else:
                    yield tracker, thread_id, 'Web', frame
            else:
                active_thread = threading._active.get(thread_id)
                if active_thread is not None and active_thread.getName().startswith('TingYun-'):
                    yield None, thread_id, 'Agent', frame
                else:
                    yield None, thread_id, 'Other', frame

    def drop_tracker(self, tracker):
        """
        """
        thread_id = tracker.thread_id

        if thread_id not in self._cache:
            RuntimeError('no active tracker')

        current = self._cache.get(thread_id, None)
        if tracker != current:
            raise RuntimeError('not the current tracker')

        del self._cache[thread_id]

    def current_tracker(self):
        """Return the tracker object if one exists for the currently executing thread.
        """
        thread_id = self.current_thread_id()
        
        return self._cache.get(thread_id, None)

_knapsack = Knapsack()


def knapsack():
    """
    """
    return _knapsack



"""this module implement the time tracer for tracker
"""

import time
import logging


console = logging.getLogger(__name__)


class Timer(object):
    """
    """
    node = None

    def __init__(self, tracker):
        self.tracker = tracker
        self.children = []
        self.start_time = 0.0
        self.end_time = 0.0
        self.duration = 0.0
        self.exclusive = 0.0
        
    def __enter__(self):
        """
        """
        if not self.tracker:
            console.debug("tracker is %s.return now.", self.tracker)
            return self

        # check the base node is set or not
        parent_node = self.tracker.current_node()
        if not parent_node or parent_node.terminal_node():
            self.tracker = None
            return parent_node

        self.start_time = time.time()
        self.tracker.push_node(self)
        
        return self
        
    def __exit__(self, exc, value, tb):
        """
        """
        if not self.tracker:
            return self.tracker

        self.end_time = time.time()
        if self.end_time < self.start_time:
            self.end_time = self.start_time
            console.warn("end time is less than start time.")

        self.duration = int((self.end_time - self.start_time) * 1000)
        self.exclusive += self.duration
        
        # pop the self and return parent node
        parent_node = self.tracker.pop_node(self)

        self.finalize_data()
        current_node = self.create_node()
        
        if current_node:
            self.tracker.process_database_node(current_node)
            parent_node.process_child(current_node)
            
        self.tracker = None
        
    def create_node(self):
        if self.node:
            return self.node(**dict((k, self.__dict__[k]) for k in self.node._fields))

        return self

    def finalize_data(self):
        pass
    
    def process_child(self, node):
        self.children.append(node)
        self.exclusive -= node.duration
        
    def terminal_node(self):
        return False




"""this module is implement some wrapper for trace the tornado web module

"""

import logging
import sys

from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.ammunition.function_tracker import FunctionTracker
from tingyun.armoury.ammunition.tornado_tracker.httpserver import generate_tracer, stop_request_tracer
from tingyun.armoury.ammunition.tornado_tracker.httpserver import setup_func_for_async_trace
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.logistics.basic_wrapper import FunctionWrapper
from tingyun.logistics.object_name import callable_name


console = logging.getLogger(__name__)


def trace_request_exception(wrapped):
    """
    :param wrapped:
    :return:
    """
    def wrapper(wrapped, instance, args, kwargs):
        def parse_args(e, *args, **kwargs):
            return e

        tracer = getattr(instance.request, "_self_tracer", None)
        if tracer:
            e = parse_args(*args, **kwargs)
            tracer.http_status = getattr(e, 'status_code', 500)
            tracer.record_exception(*sys.exc_info())

        return wrapped(*args, **kwargs)

    return FunctionWrapper(wrapped, wrapper)


def trace_wsgi_app_entrance(wrapped):
    """flowing situations maybe faced.

        1. the tornado application transferred to run with WSGI server
        2. WSGI application run in tornado server.
        3. tornado application run in tornado server.
    """
    def parse_param(request, *args, **kwargs):
            return request, args, kwargs

    def trace_app_without_body(wrapped, instance, request, args, kwargs):
        """
        :return:
        """
        tracker = generate_tracer(request)
        if not tracker:
            return wrapped(*args, **kwargs)

        try:
            # we should use `FunctionWrapper` at here.because the Connection._on_finished will be executed in the
            # wrapped func. use `FunctionWrapper` will cause errors.
            ret = wrapped(*args, **kwargs)
        except:
            stop_request_tracer(request, *sys.exc_info(), segment="App.__call__.exception")
            raise
        else:
            if request._self_request_finished:
                return ret

            request._self_request_finished = True
            stop_request_tracer(request, segment='finish-app-call')
            return ret

    def wrapper(wrapped, instance, args, kwargs):
        """tracker will be put into the thread for other plugins trace in user code. such as mysql/memcached etc.
        :return:
        """
        request, _args, _kwargs = parse_param(*args, **kwargs)
        async_tracker = getattr(request, '_self_tracer', None)
        thread_tracker = current_tracker()

        # situation 1 & 2. just do nothing because the entrance of the wsgi app was detected.
        if not async_tracker and thread_tracker:
            # console.info("has not self tracer.%s, thread tracer %s", async_tracker, thread_tracker)
            return wrapped(*args, **kwargs)

        # situation 3. but no request body send by the client, at this case, the tracker was not set into the request
        # in HTTPConnection._on_headers
        if not async_tracker:
            return trace_app_without_body(wrapped, instance, request, args, kwargs)

        # situation 3. client send the header & body content to server for requesting data.
        return wrapped(*args, **kwargs)

    return FunctionWrapper(wrapped, wrapper)


def trace_request_execute(wrapped):
    """
    """
    def wrapper(wrapped, instance, args, kwargs):
        """
        """
        request = instance.request
        tracer = getattr(request, "_self_tracer", None)
        if not tracer:
            return wrapped(*args, **kwargs)

        name = callable_name(getattr(instance, instance.request.method.lower()))
        tracer.set_tracker_name(name, priority=3)

        with FunctionTracker(tracer, callable_name(wrapped)):
            thread_tracer = current_tracker()
            if thread_tracer:
                console.info("Ignore user code metric trace for current request, this will maybe cause lack of some "
                             "metric data. if this continues, please report to us, thank u.")
                return wrapped(*args, **kwargs)

            # at here, we get the user code. we just put the tracer to the thread for non-framework plugin use.after
            # this, we drop it again. and ignore all of the errors occurred.
            try:
                tracer.save_tracker()
            except Exception as _:
                console.error("Errors, when the tracer put into the thread, this should not occurred, if this "
                              "continues, please report to us, thank u")
            ret = wrapped(*args, **kwargs)
            try:
                tracer.drop_tracker()
            except Exception as _:
                console.error("Errors, when the tracer dropped from the thread, this should not occurred, if this "
                              "continues, please report to us, thank u")

            return ret

    return FunctionWrapper(wrapped, wrapper)


def trace_request_init(wrapped):
    """hook the __init__ just for every class instance method.
    """
    def trace_user_func(wrapped):
        """
        """
        def user_func_wrapper(wrapped, instance, args, kwargs):
            tracer = getattr(instance.request, "_self_tracer", None)
            if not tracer:
                # console.info("ignore trace func %s", callable_name(wrapped))
                return wrapped(*args, **kwargs)

            with FunctionTracker(tracer, callable_name(wrapped)):
                # console.info("---------------trace func %s", callable_name(wrapped))
                return wrapped(*args, **kwargs)

        return FunctionWrapper(wrapped, user_func_wrapper)

    def wrapper(wrapped, instance, args, kwargs):
        """
        """
        instance.prepare = trace_user_func(instance.prepare)
        instance.on_finish = trace_user_func(instance.on_finish)
        instance.on_connection_close = trace_user_func(instance.on_connection_close)
        instance.initialize = trace_user_func(instance.initialize)

        for name in instance.SUPPORTED_METHODS:
            name = name.lower()
            if hasattr(instance, name):
                setattr(instance, name, trace_user_func(getattr(instance, name)))

        return wrapped(*args, **kwargs)

    return FunctionWrapper(wrapped, wrapper)




"""The detect module follow the protocol: https://www.python.org/dev/peps/pep-0249/

"""

import logging

from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.ammunition.database_tracker import DatabaseTracker
from tingyun.armoury.ammunition.function_tracker import FunctionTracker, wrap_function_trace


# module_name: standard db name for data exchange protocol.
db_name_module = {
    "MySQLdb": "MySQL", "pymysql": "MySQL", "oursql": "MySQL",
    "cx_Oracle": "Oracle",
    "psycopg2": "PostgreSQL", "psycopg2ct": "PostgreSQL", "psycopg2cffi": "PostgreSQL",
    "pyodbc": "ODBC"
    }
console = logging.getLogger(__name__)


def detect(module):
    """ more interface description about dbapi2 in: https://www.python.org/dev/peps/pep-0249/
    :param module:
    :return:
    """
    dbtype = db_name_module.get(module.__name__, "dbapi2")

    class TingYunCursor(object):

        def __init__(self, cursor, cursor_params=None, connect_params=None):
            """
            :param cursor:
            :param cursor_params:
            :param connect_params:
            :return:
            """
            object.__setattr__(self, 'cursor', cursor)
            object.__setattr__(self, 'cursor_params', cursor_params)
            object.__setattr__(self, 'connect_params', connect_params)

        def __setattr__(self, name, value):
            setattr(self.cursor, name, value)

        def __getattr__(self, name):
            return getattr(self.cursor, name)

        def __iter__(self):
            return iter(self.cursor)

        def fetchone(self, *args, **kwargs):
            """we do not capture the metric of execute result. this is small time used
            :args:
            :kwargs:
            :return:
            """
            return self.cursor.fetchone(*args, **kwargs)

        def fetchmany(self, *args, **kwargs):
            """we do not capture the metric of execute result. this is small time used
            :args:
            :kwargs:
            :return:
            """
            return self.cursor.fetchmany(*args, **kwargs)

        def fetchall(self, *args, **kwargs):
            """this operation maybe spend more time. this is small time used
            and the sql was executed. we can not take it
            :args:
            :kwargs:
            :return:
            """
            return self.cursor.fetchall(*args, **kwargs)

        def execute(self, sql, *args, **kwargs):
            """
            :param sql:
            :param args:
            :param kwargs:
            :return:
            """
            tracker = current_tracker()
            if not tracker:
                return self.cursor.execute(sql, *args, **kwargs)

            with DatabaseTracker(tracker, sql, dbtype, module, self.connect_params, self.cursor_params,
                                 (args, kwargs)):
                return self.cursor.execute(sql, *args, **kwargs)

        def executemany(self, sql, *args, **kwargs):
            """
            :param sql:
            :param args:
            :param kwargs:
            :return:
            """
            tracker = current_tracker()
            if not tracker:
                return self.cursor.executemany(sql, *args, **kwargs)

            with DatabaseTracker(tracker, sql, dbtype, module):
                return self.cursor.executemany(sql, *args, **kwargs)

        def callproc(self, procname, *args, **kwargs):
            """
            :param procname:
            :param args:
            :param kwargs:
            :return:
            """
            tracker = current_tracker()
            if not tracker:
                return self.cursor.callproc(procname, *args, **kwargs)

            with DatabaseTracker(tracker, 'CALL %s' % procname, dbtype, module):
                return self.cursor.callproc(procname, *args, **kwargs)

    class TingYunConnection(object):

        def __init__(self, connection, params=None):
            """
            :param connection:
            :param params:
            :return:
            """
            object.__setattr__(self, 'connection', connection)
            object.__setattr__(self, 'params', params)

        def __setattr__(self, name, value):
            setattr(self.connection, name, value)

        def __getattr__(self, name):
            return getattr(self.connection, name)

        def close(self, *args, **kwargs):
            """
            :param args:
            :param kwargs:
            :return:
            """
            return self.connection.close(*args, **kwargs)

        def cursor(self, *args, **kwargs):
            """
            :param args:
            :param kwargs:
            :return:
            """
            return TingYunCursor(self.connection.cursor(*args, **kwargs), (args, kwargs), self.params)

        def commit(self):
            """
            :return:
            """
            tracker = current_tracker()
            if not tracker:
                return self.connection.commit()

            with DatabaseTracker(tracker, 'COMMIT', dbtype, module):
                return self.connection.commit()

        def rollback(self):
            """
            :return:
            """
            tracker = current_tracker()
            if not tracker:
                return self.connection.rollback()

            with DatabaseTracker(tracker, 'ROLLBACK', dbtype, module):
                return self.connection.rollback()

    class DatabaseWrapper(object):

        def __init__(self, connect):
            self.connect = connect
            self.connect_instance = None

        def __call__(self, *args, **kwargs):
            """when module call the connect method. the wrapper will callback __call__ instead.
            :param args:
            :param kwargs:
            :return:
            """
            self.connect_instance = self.connect(*args, **kwargs)

            return TingYunConnection(self.connect_instance, (args, kwargs))

    # Check if module is already wrapped
    if hasattr(module, '_self_dbapi2_wrapped'):
        return

    wrap_function_trace(module, 'connect', name='%s.connect' % dbtype, group='Database')
    module.connect = DatabaseWrapper(module.connect)
    module._self_dbapi2_wrapped = True




"""

"""

from collections import namedtuple
from tingyun.logistics.attribution import TimeMetric, node_start_time, node_end_time

_ExternalNode = namedtuple('_ExternalNode', ['library', 'url', 'children', 'start_time', 'end_time', 'protocol',
                                             'duration', 'exclusive'])


class ExternalNode(_ExternalNode):
    """

    """
    def time_metrics(self, root, parent):
        """
        :param root: the top node of the tracker
        :param parent: parent node.
        :return:
        """
        def extend_metric(metric_name):
            if not root.trace_data:
                return metric_name

            return "%s|%s|%s" % (metric_name, root.trace_data.get('id'), root.trace_data.get('action'))

        name = 'GENERAL/External/NULL/All'
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        name = "GENERAL/External/NULL/AllWeb"
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        name = 'External/%s/%s' % (self.url.replace("/", "%2F"), root.name)
        yield TimeMetric(name=extend_metric(name), scope=root.path, duration=self.duration, exclusive=self.exclusive)

        if root.trace_data:
            # for cross trace.
            trace_data = root.trace_data
            name = 'ExternalTransaction/NULL/%s' % trace_data.get("id")
            yield TimeMetric(name=extend_metric(name), scope=root.path, duration=self.duration, exclusive=self.exclusive)

            name = 'ExternalTransaction/%s/%s' % (self.protocol, trace_data.get("id"))
            yield TimeMetric(name=extend_metric(name), scope=root.path, duration=self.duration, exclusive=self.exclusive)

            name = 'ExternalTransaction/%s:sync/%s' % (self.protocol, trace_data.get("id"))
            yield TimeMetric(name=extend_metric(name), scope=root.path, duration=self.duration, exclusive=self.exclusive)

        name = 'GENERAL/External/%s/%s' % (self.url.replace("/", "%2F"), root.name)
        yield TimeMetric(name=extend_metric(name), scope=root.path, duration=self.duration, exclusive=self.exclusive)

    def trace_node(self, root):
        """
        :param root: the root node of the tracker
        :return:
        """
        params = {}
        children = []
        call_count = 1
        class_name = ""
        method_name = root.name
        call_url = self.url
        root.trace_node_count += 1
        start_time = node_start_time(root, self)
        end_time = node_end_time(root, self)
        metric_name = 'External/%s/%s' % (self.url.replace("/", "%2F"), root.name)

        if root.trace_id and root.trace_data:
            params['txId'] = root.trace_id
            params['txData'] = root.trace_data

        return [start_time, end_time, metric_name, call_url, call_count, class_name, method_name, params, children]




"""define a detector for tornado ioloop
"""

import sys
import logging
from tingyun.armoury.ammunition.tornado_4.utils import finish_tracker, record_exception, obtain_current_tracker
from tingyun.armoury.ammunition.tornado_4.utils import current_thread_id, TrackerTransferContext
from tingyun.logistics.basic_wrapper import wrap_function_wrapper

console = logging.getLogger(__name__)


def trace_run_callback(wrapped, instance, args, kwargs):
    """
    """
    def _get_actually_callback(callback, *args, **kwargs):
        try:
            return callback.func
        except AttributeError:
            return None

    callback = _get_actually_callback(*args, **kwargs)
    tracker = getattr(callback, '_self_tracker', None)
    ret = wrapped(*args, **kwargs)

    if tracker:
        tracker._ref_count -= 1
        finish_tracker(tracker)

    return ret


def trace_handle_callback_exception(wrapped, instance, args, kwargs):
    """
    """
    record_exception(sys.exc_info())

    return wrapped(*args, **kwargs)


def _increment_ref_count(callback, wrapped, instance, args, kwargs):
    tracker = obtain_current_tracker()

    if hasattr(callback, '_self_tracker'):

        if callback._self_tracker is not None:
            if current_thread_id() != callback._self_tracker.thread_id:
                # Callback being added not in the main thread; ignore.

                # Since we are not incrementing the counter for this callback,
                # we need to remove the tracker from the callback, so it
                # doesn't get decremented either.
                callback._self_tracker = None
                return wrapped(*args, **kwargs)

        if tracker is not callback._self_tracker:
            console.warning("Attempt to add callback to ioloop with different tracer attached than in the cache."
                            "if this continue, Please report to us.")
            callback._self_tracker = None
            return wrapped(*args, **kwargs)

    if tracker is None:
        return wrapped(*args, **kwargs)

    tracker._ref_count += 1

    return wrapped(*args, **kwargs)


def trace_add_callback(wrapped, instance, args, kwargs):
    """
    """
    def _get_actually_callback(callback, *args, **kwargs):
        return callback

    callback = _get_actually_callback(*args, **kwargs)
    return _increment_ref_count(callback, wrapped, instance, args, kwargs)


def trace_call_at(wrapped, instance, args, kwargs):
    """
    """
    with TrackerTransferContext(None):
        return wrapped(*args, **kwargs)


def trace_add_handler(wrapped, instance, args, kwargs):
    """
    """
    with TrackerTransferContext(None):
        return wrapped(*args, **kwargs)


def detect_ioloop(module):
    """
    :param module:
    :return:
    """
    wrap_function_wrapper(module, 'IOLoop._run_callback', trace_run_callback)
    wrap_function_wrapper(module, 'IOLoop.handle_callback_exception', trace_handle_callback_exception)
    wrap_function_wrapper(module, 'PollIOLoop.add_callback', trace_add_callback)
    wrap_function_wrapper(module, 'PollIOLoop.call_at', trace_call_at)
    wrap_function_wrapper(module, 'PollIOLoop.add_handler', trace_add_handler)



from __future__ import print_function

import os
import sys
from tingyun.logistics.exceptions import CommandlineParametersException
from tingyun.logistics.mapper import ENV_CONFIG_FILE
from tingyun.config.start_log import log_message
from tingyun import __file__ as root_dir


class Command(object):
    """
    """
    def __init__(self):
        """
        """
        self.name = "run-program"  # order format with commandline
        self.options = 'command [parameters]'
        self.description = "Executes the command with parameters."

    def execute(self, args):
        """
        :param args:
        :return:
        """
        if 0 == len(args):
            raise CommandlineParametersException()

        log_message('-------------get in bootstrap--------------')
        log_message('TingYun Admin Script (%s)', __file__)
        log_message('working_directory = %r', os.getcwd())
        log_message('current_command = %r', sys.argv)
        log_message('sys.prefix = %r', os.path.normpath(sys.prefix))

        log_message('sys.executable = %r', sys.executable)
        log_message('sys.flags = %r', sys.flags)
        log_message('sys.path = %r', sys.path)

        boot_directory = os.path.join(os.path.dirname(root_dir), 'flashpoint')
        log_message('boot_directory = %r', boot_directory)

        # before change the pythonpath. we should load the exist, then add myself into
        final_python_path = boot_directory
        if 'PYTHONPATH' in os.environ:
            path = os.environ['PYTHONPATH'].split(os.path.pathsep)
            if boot_directory not in path:
                final_python_path = "%s%s%s" % (boot_directory, os.path.pathsep, os.environ['PYTHONPATH'])

        log_message('python_path = %r', final_python_path)
        os.environ['PYTHONPATH'] = final_python_path

        # deal the program exe as a system command, and change to full system path
        program_exe_path = args[0]
        if not os.path.dirname(program_exe_path):
            program_target_path = os.environ.get('PATH', '').split(os.path.pathsep)
            log_message("get the path from env:%s", program_target_path)
            for path in program_target_path:
                program_exe_path_tmp = os.path.join(path, program_exe_path)
                if os.path.exists(program_exe_path_tmp) and os.access(program_exe_path_tmp, os.X_OK):
                    program_exe_path = program_exe_path_tmp
                    log_message("match the program exe: %s", program_exe_path)
                    break

        log_message('program_exe_path = %r', program_exe_path)
        log_message('execl_arguments = %r', [program_exe_path] + args)

        os.execl(program_exe_path, *args)

        # can not check the config before execute user command.
        # because may not detect the environment variable in situation that environment variable is set in shell
        config_file = os.environ.get(ENV_CONFIG_FILE, None)
        if config_file is None:
            print('''
            *************************Warning***********************

                Agent does not obtain the environment
            variable `%s` for config file. agent did not work.
                Please set the config file first, and then restart your
            application with agent

            *******************************************************
            ''' % ENV_CONFIG_FILE)
            return
        elif os.path.isfile(config_file):
            print('''
            *************************Warning***********************

                The config file that environment variable point to is not
            exist. agent did not work.
                Please set the config file first, and then restart your
            application with agent

            *******************************************************
            ''')
            return
        else:
            readable = True
            try:
                with open(config_file, 'r'):
                    pass
            except Exception as _:
                readable = False

            if not readable:
                print('''
                *************************Warning***********************

                    The config file `%s`
                is not readable. agent did not work.
                    Please set the config file first, and then restart your
                application with agent

                *******************************************************
                ''' % config_file)




"""

"""
import sys
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.ammunition.external_tracker import wrap_external_trace, parse_parameters
from tingyun.armoury.ammunition.external_tracker import OTHER_ERROR_CODE, ILLEGAL_RESPONSE_ERROR_CODE

try:
    from urllib import ContentTooShortError
except ImportError:
    class ContentTooShortError(Exception):
        pass


def wrap_exception(wrapped, url, parameters, *args, **kwargs):
    """
    """

    tracker = current_tracker()
    if tracker is None:
        return wrapped(*args, **kwargs)

    url = url if not callable(url) else url(*args, **kwargs)
    params = parameters if not callable(parameters) else parameters(*args, **kwargs)

    http_status = 500
    error_code = 0
    params = parse_parameters(url, params)

    try:
        rtv = wrapped(*args, **kwargs)
    except ILLEGAL_RESPONSE_ERROR_CODE:
        error_code = ILLEGAL_RESPONSE_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params)
        raise
    except Exception:
        error_code = OTHER_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params)
        raise

    try:
        http_status = rtv.getcode()
        if http_status is not None and int(http_status) != 200:
            tracker.record_external_error(url, error_code, int(http_status), sys.exc_info(), params)
    except Exception:
        pass

    return rtv


def detect(module):

    def urllib_url(instance, fullurl, *args, **kwargs):
        return fullurl

    if hasattr(module, 'URLopener'):
        wrap_external_trace(module, 'URLopener.open', 'urllib', urllib_url, exception_wrapper=wrap_exception)



from __future__ import print_function
import os
import time

startup_debug = os.environ.get('TING_YUN_STARTUP_DEBUG', 'off').lower() in ('on', 'true', '1')
log_file = "/tmp/tingyun_bootstrap.log"
log_file_handler = None


def log_message(text, *args):
    global startup_debug
    if startup_debug:
        text = text % args
        timestamp = time.strftime('%m-%d %T', time.localtime())
        print('TingYun: %s (%d) - %s' % (timestamp, os.getpid(), text))

        return True
    
    return False


def log_bootstrap(msg, close=False):
    """
    :param msg:
    :return:
    """
    global log_file_handler

    if not startup_debug:
        return

    if close and log_file_handler is not None:
        log_data = "%s %s %s\n" % (time.strftime("%Y-%m-%d %H:%M:%S ",  time.localtime()), os.getpid(), str(msg))
        log_file_handler.write(log_data)
        log_file_handler.write("close the log file....\n")
        log_file_handler.close()
        log_file_handler = None
        return

    if close and log_file_handler is None:
        log_file_handler = open(log_file, mode='a')
        log_data = "%s %s %s\n" % (time.strftime("%Y-%m-%d %H:%M:%S ",  time.localtime()), os.getpid(), str(msg))
        log_file_handler.write(log_data)
        log_file_handler.write("close the log file....\n")
        log_file_handler.close()
        log_file_handler = None
        return

    if not close and log_file_handler is None:
        log_file_handler = open(log_file, mode='a')
        log_file_handler.write("open the log file....\n")

    log_data = "%s %s %s\n" % (time.strftime("%Y-%m-%d %H:%M:%S ",  time.localtime()), os.getpid(), str(msg))
    log_file_handler.write(log_data)




"""define some wrapper to wrapt some method in tornado http server
"""

import logging
from tingyun.logistics.basic_wrapper import wrap_function_wrapper
from tingyun.armoury.ammunition.tornado_4.utils import finish_tracker

console = logging.getLogger(__name__)


def trace_tracker_export(wrapped, instance, args, kwargs):
    """
    """
    # the request maybe exist two situations in httpserver delegate
    request = instance.delegate.request if instance.delegate else instance.request

    if not request:
        console.warning("No request got in _ServerRequestAdapter object. this should not be happen. if this continue."
                        " please report us.")
        return wrapped(*args, **kwargs)

    tracker = getattr(request, "_self_tracker", None)
    if not tracker:
        return wrapped(*args, **kwargs)

    try:
        wrapped(*args, **kwargs)
    finally:
        setattr(tracker, "_can_finalize", True)
        finish_tracker(tracker)


def detect_tracker_export(module):
    """detect the export of the tracker.
    :param module: tornado httpserver
    :return:
    """
    wrap_function_wrapper(module, "_ServerRequestAdapter.on_connection_close", trace_tracker_export)
    wrap_function_wrapper(module, "_ServerRequestAdapter.finish", trace_tracker_export)




"""
"""

import logging

from tingyun.armoury.ammunition.external_tracker import wrap_external_trace
from tingyun.armoury.ammunition.tornado_tracker.httpserver import connection_on_headers_wrapper
from tingyun.armoury.ammunition.tornado_tracker.httpserver import connect_on_request_body_wrapper
from tingyun.armoury.ammunition.tornado_tracker.httpserver import connection_finish_request_wrapper
from tingyun.armoury.ammunition.tornado_tracker.httpserver import iostream_close_callback_wrapper
from tingyun.armoury.ammunition.tornado_tracker.web import trace_request_exception, trace_wsgi_app_entrance
from tingyun.armoury.ammunition.tornado_tracker.web import trace_request_execute, trace_request_init


console = logging.getLogger(__name__)


def detect_wsgi_server_entrance(module):
    """
    :param module:
    :return:
    """
    # todo: we give up to detect the WSGIAdapter(v4.x)/WSGIApplication/WSGIContainer just because:
    # todo: 1. WSGIAdapter & WSGIApplication, we have detect the tornado application.__call__, so there is no need to
    # todo:    detect again.
    # todo: 2. WSGIContainer, give up to detect the entrance will suffer one situation, which can not capture the
    # todo:    request metric if tornado server run a WSGI app that we are not support.
    return
    # import tornado
    #
    # version = getattr(tornado, "version", "xx")
    #
    # # new feature in tornado 4.x
    # if hasattr(module, 'WSGIAdapter'):
    #     wsgi_application_wrapper(module.WSGIAdapter, "__call__", ("Tornado", version))
    #
    # # equivalent to WSGIAdapter call
    # if hasattr(module, "WSGIApplication"):
    #     # wsgi_application_wrapper(module.WSGIApplication, "__call__", ("Tornado", version))
    #     module.Application.__call__ = trace_wsgi_app_entrance(module.Application.__call__)
    #
    # def wsgi_container_wrapper(*args, **kwargs):
    #     """
    #     """
    #     def instance_parameters(wsgi_application, *args, **kwargs):
    #         return wsgi_application, args, kwargs
    #
    #     application, _args, _kwargs = instance_parameters(*args, **kwargs)
    #     application = wsgi_app_wrapper_entrance(application)
    #
    #     _args = (application, ) + _args
    #     return _args, _kwargs
    #
    # # Makes a WSGI-compatible function runnable on Tornado's HTTP server.
    # if hasattr(module, "WSGIContainer"):
    #     trace_in_function(module.WSGIContainer, "__init__", wsgi_container_wrapper)


def detect_wsgi_app_entrance(module):
    """
    """
    # in tornado 4.x.x, the application.__call__ will not be called directly.
    module.Application.__call__ = trace_wsgi_app_entrance(module.Application.__call__)

    if hasattr(module, "RequestHandler"):
        module.RequestHandler.__init__ = trace_request_init(module.RequestHandler.__init__)
        module.RequestHandler._handle_request_exception = trace_request_exception(module.RequestHandler._handle_request_exception)
        module.RequestHandler._execute = trace_request_execute(module.RequestHandler._execute)

    # for tornado 4.x
    # if hasattr(module, "_RequestDispatcher"):
    #     module._RequestDispatcher.headers_received = connection_on_headers_wrapper(module._RequestDispatcher.headers_received)
    #     module._RequestDispatcher.data_received = connect_on_request_body_wrapper(module._RequestDispatcher.data_received)
    #     module._RequestDispatcher.finish = connection_finish_request_wrapper(module._RequestDispatcher.finish)


def detect_tornado_main_process(module):
    """all of the data handled in HTTPConnection class, include build header/body/finish
    :param module:
    :return:
    """
    # for tornado 3.x
    if hasattr(module, "HTTPConnection"):
        module.HTTPConnection._on_headers = connection_on_headers_wrapper(module.HTTPConnection._on_headers)
        module.HTTPConnection._on_request_body = connect_on_request_body_wrapper(module.HTTPConnection._on_request_body)
        module.HTTPConnection._finish_request = connection_finish_request_wrapper(module.HTTPConnection._finish_request)


def detect_iostream(module):
    """
    :param module:
    :return:
    """

    if hasattr(module, "BaseIOStream"):
        module.BaseIOStream._maybe_run_close_callback = iostream_close_callback_wrapper(module.BaseIOStream._maybe_run_close_callback)


def detect_simple_httpclient(module):
    """
    :param module:
    :return:
    """
    def parse_url(instance, request, *args, **kwargs):
        return request

    wrap_external_trace(module, 'SimpleAsyncHTTPClient.fetch', 'simple_httpclient', parse_url)


def detect_curl_httpclient(module):
    """
    :param module:
    :return:
    """
    def parse_url(instance, request, *args, **kwargs):
        return request

    wrap_external_trace(module, 'CurlAsyncHTTPClient.fetch', 'curl_httpclient', parse_url)



"""
"""

from tingyun.config.settings import global_settings

tornado_version = ""

try:
    import tornado
    tornado_version = tornado.version
except Exception:
    pass


def defined_repertoire():
    """
    :return:
    """
    hookers = {
        "memcached": [
            {"target": "memcache", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_memcached'},
        ],

        # mysql db
        "mysql": [
            {"target": "MySQLdb", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],
        "pymysql": [
            {"target": "pymysql", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],
        "oursql": [
            {"target": "oursql", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],

        # oracle
        "oracle": [
            {"target": "cx_Oracle", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],

        # postgres SQL
        "postgresql": [
            {"target": "postgresql.interface.proboscis.dbapi2", 'hook_func': 'detect',
             'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],
        # postgres SQL
        "psycopg2": [
            {"target": "psycopg2", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],
        "psycopg2ct": [
            {"target": "psycopg2ct", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],
        "psycopg2cffi": [
            {"target": "psycopg2cffi", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],

        # ODBC A Python DB API 2 module for ODBC
        "pyodbc": [
            {"target": "pyodbc", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.database_dbapi2'},
        ],

        # nosql mongodb
        "mongodb": [
            {"target": "pymongo.mongo_client", 'hook_func': 'detect_mongo_client',
             'hook_module': 'tingyun.armoury.database_mongo'},
            {"target": "pymongo.connection", 'hook_func': 'detect_connection',
             'hook_module': 'tingyun.armoury.database_mongo'},
            {"target": "pymongo.collection", 'hook_func': 'detect_collection',
             'hook_module': 'tingyun.armoury.database_mongo'},
        ],

        # nosql redis
        "redis": [
            {"target": "redis.connection", 'hook_func': 'detect_connection',
             'hook_module': 'tingyun.armoury.database_redis'},
            {"target": "redis.client", 'hook_func': 'detect_client_operation',
             'hook_module': 'tingyun.armoury.database_redis'},
        ],

        # external call
        "urllib": [
            {"target": "urllib", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.external_urllib'},
        ],
        "urllib2": [
            {"target": "urllib2", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.external_urllib2'},
        ],
        # v1.6-v1.14
        "urllib3": [
            {"target": "urllib3.poolmanager", 'hook_func': 'detect', 'hook_module': 'tingyun.armoury.external_urllib3'},
        ],
        # 0.8.0-0.9.3
        "thrift": [
            {"target": "thrift.transport.TSocket", 'hook_func': 'detect_tsocket',
             'hook_module': 'tingyun.armoury.external_thrift'},
            {"target": "thrift.transport.TSSLSocket", 'hook_func': 'detect_tsslsocket',
             'hook_module': 'tingyun.armoury.external_thrift'},
        ],

        #2.0.0-2.10.0
        "requests": [
            {"target": "requests.api", 'hook_func': 'detect_requests_request',
             'hook_module': 'tingyun.armoury.external_requests'},
            {"target": "requests.sessions", 'hook_func': 'detect_requests_sessions',
             'hook_module': 'tingyun.armoury.external_requests'},
        ],
        # 0.7.5-0.9.2
        "httplib2": [
            {"target": "httplib2", 'hook_func': 'detect_httplib2_http',
             'hook_module': 'tingyun.armoury.external_httplib2'},
            {"target": "httplib2", 'hook_func': 'detect_http_connect_with_timeout',
             'hook_module': 'tingyun.armoury.external_httplib2'},
            {"target": "httplib2", 'hook_func': 'detect_https_connect_with_timeout',
             'hook_module': 'tingyun.armoury.external_httplib2'},
        ],

        # django, this weapon must not be removed.
        "django": [
            {"target": "django.core.handlers.base", 'hook_func': 'detect_middleware',
             'hook_module': 'tingyun.armoury.framework_django'},
            {"target": "django.core.handlers.wsgi", 'hook_func': 'detect_wsgi_entrance',
             'hook_module': 'tingyun.armoury.framework_django'},
            {"target": "django.core.urlresolvers", 'hook_func': 'detect_urlresolvers',
             'hook_module': 'tingyun.armoury.framework_django'},
            {"target": "django.views.generic.base", 'hook_func': 'detect_views_dispatch',
             'hook_module': 'tingyun.armoury.framework_django'},

            {"target": "django.template.loader_tags", 'hook_func': 'detect_template_block_render',
             'hook_module': 'tingyun.armoury.framework_django'},
            {"target": "django.template.base", 'hook_func': 'detect_django_template',
             'hook_module': 'tingyun.armoury.framework_django'},

            {"target": "django.http.multipartparser", 'hook_func': 'detect_http_multipartparser',
             'hook_module': 'tingyun.armoury.framework_django'},

            {"target": "django.core.mail", 'hook_func': 'detect_core_mail',
             'hook_module': 'tingyun.armoury.framework_django'},
            {"target": "django.core.mail.message", 'hook_func': 'detect_core_mail_message',
             'hook_module': 'tingyun.armoury.framework_django'},
        ],

        # django-piston, 0.2.2.1-0.2.3
        "piston": [
            {"target": "piston.resource", 'hook_func': 'detect_piston_resource',
             'hook_module': 'tingyun.armoury.ammunition.piston'},
        ],

        # flask, 0.6-1.0
        "flask": [
            {"target": "flask.app", 'hook_func': 'detect_wsgi_entrance',
             'hook_module': 'tingyun.armoury.framework_flask'},
            {"target": "flask.app", 'hook_func': 'detect_app_entrance',
             'hook_module': 'tingyun.armoury.framework_flask'},
            {"target": "flask.blueprints", 'hook_func': 'detect_app_blueprint_entrance',
             'hook_module': 'tingyun.armoury.framework_flask'},
            {"target": "flask.templating", 'hook_func': 'detect_templates',
             'hook_module': 'tingyun.armoury.framework_flask'},
        ],

        # jinja2 2.3-2.8
        'jinja2': [
            {"target": "jinja2.loaders", 'hook_func': 'detect_template_loader',
             'hook_module': 'tingyun.armoury.template_jinja2'},
            {"target": "jinja2.environment", 'hook_func': 'detect_jinja2',
             'hook_module': 'tingyun.armoury.template_jinja2'},
            ],

        # version 2.8.1-2.12.x
        'web2py': [
            {"target": "gluon.main", 'hook_func': 'detect_wsgi_entrance',
             'hook_module': 'tingyun.armoury.framework_web2py'},
            {"target": "gluon.compileapp", 'hook_func': 'detect_compileapp',
             'hook_module': 'tingyun.armoury.framework_web2py'},
            {"target": "gluon.template", 'hook_func': 'detect_template',
             'hook_module': 'tingyun.armoury.framework_web2py'},
        ],

        # version 0.3.x
        'webpy': [
            {"target": "web.application", 'hook_func': 'detect_wsgi_entrance',
             'hook_module': 'tingyun.armoury.framework_webpy'},

            {"target": "web.application", 'hook_func': 'detect_application',
             'hook_module': 'tingyun.armoury.framework_webpy'},
        ],

        # version 3.x.x
        "tornado3": [
            {"target": "tornado.wsgi", 'hook_func': 'detect_wsgi_server_entrance',
             'hook_module': 'tingyun.armoury.framework_tornado'},
            {"target": "tornado.httpserver", 'hook_func': 'detect_tornado_main_process',
             'hook_module': 'tingyun.armoury.framework_tornado'},
            {"target": "tornado.web", 'hook_func': 'detect_wsgi_app_entrance',
             'hook_module': 'tingyun.armoury.framework_tornado'},
            {"target": "tornado.web", 'hook_func': 'detect_iostream',
             'hook_module': 'tingyun.armoury.framework_tornado'},
            {"target": "tornado.simple_httpclient", 'hook_func': 'detect_simple_httpclient',
             'hook_module': 'tingyun.armoury.framework_tornado'},
            {"target": "tornado.curl_httpclient", 'hook_func': 'detect_curl_httpclient',
             'hook_module': 'tingyun.armoury.framework_tornado'},
        ] if "3." in tornado_version else [],

        # version 4.x.x
        # when tornado only use as wsgi application, do not detect tornado component
        "tornado4": [
            {"target": "tornado.web", 'hook_func': 'detect_handlers',
             'hook_module': 'tingyun.armoury.ammunition.tornado_4.web'},

            {"target": "tornado.httputil", 'hook_func': 'detect_tracker_entrance',
             'hook_module': 'tingyun.armoury.ammunition.tornado_4.http_util'},
            {"target": "tornado.httpserver", 'hook_func': 'detect_tracker_export',
             'hook_module': 'tingyun.armoury.ammunition.tornado_4.http_server'},
            {"target": "tornado.httpclient", 'hook_func': 'detect_http_client',
             'hook_module': 'tingyun.armoury.ammunition.tornado_4.http_client'},
            {"target": "tornado.gen", 'hook_func': 'detect_gen',
             'hook_module': 'tingyun.armoury.ammunition.tornado_4.gen'},
            {"target": "tornado.ioloop", 'hook_func': 'detect_ioloop',
             'hook_module': 'tingyun.armoury.ammunition.tornado_4.ioloop'},
            {"target": "tornado.concurrent", 'hook_func': 'detect_concurrent',
             'hook_module': 'tingyun.armoury.ammunition.tornado_4.concurrent'},
            {"target": "concurrent.futures", 'hook_func': 'detect_concurrent',
             'hook_module': 'tingyun.armoury.ammunition.tornado_4.concurrent'},
            {"target": "tornado.stack_context", 'hook_func': 'detect_stack_context',
             'hook_module': 'tingyun.armoury.ammunition.tornado_4.stack_context'},
        ]if "4." in tornado_version and not global_settings().tornado_wsgi_adapter_mode else [],

        "tornado-wsgi-adapter-mode": [
            {"target": "tornado.web", 'hook_func': 'detect_handlers',
             'hook_module': 'tingyun.armoury.ammunition.tornado_4.wsgi_mode.web'},
            {"target": "tornado.httpclient", 'hook_func': 'detect_http_client',
             'hook_module': 'tingyun.armoury.ammunition.tornado_4.wsgi_mode.http_client'},
            {"target": "tornado.wsgi", 'hook_func': 'detect_wsgi_entrance',
             'hook_module': 'tingyun.armoury.ammunition.tornado_4.wsgi_mode.wsgi'},
        ]if global_settings().tornado_wsgi_adapter_mode else [],

        # mako v0.7.x-v1.0.x
        "mako": [
            {"target": "mako.template", 'hook_func': 'detect_template',
             'hook_module': 'tingyun.armoury.template_mako'},
        ],

        # 0.13.0-1.0.2
        "gevent": [
            {"target": "gevent.pywsgi", 'hook_func': 'detect_wsgi',
             'hook_module': 'tingyun.armoury.matrix.t_gevent'},
            {"target": "gevent.wsgi", 'hook_func': 'detect_pywsgi',
             'hook_module': 'tingyun.armoury.matrix.t_gevent'},
        ],

        # openstack-nova
        'openstack-nova': [
            # nova.api.openstack.wsgi
            {"target": "nova.api.openstack.wsgi", 'hook_func': 'resource_wsgi_entrance',
             'hook_module': 'tingyun.armoury.nova.api_openstack_wsgi'},

            # nova.wsgi
            {"target": "nova.wsgi", 'hook_func': 'middleware_wsgi_entrance',
             'hook_module': 'tingyun.armoury.nova.wsgi'},
            {"target": "nova.wsgi", 'hook_func': 'router_wsgi_entrance',
             'hook_module': 'tingyun.armoury.nova.wsgi'},

            # nova.api.auth
            {"target": "nova.api.auth", 'hook_func': 'keystone_context_wsgi_entrance',
             'hook_module': 'tingyun.armoury.nova.api_auth'},

            # nova.api.ec2
            {"target": "nova.api.ec2", 'hook_func': 'ec2_keystone_wsgi_entrance',
             'hook_module': 'tingyun.armoury.nova.api_ec2'},
            {"target": "nova.api.ec2", 'hook_func': 'requestify_wsgi_entrance',
             'hook_module': 'tingyun.armoury.nova.api_ec2'},
            {"target": "nova.api.ec2", 'hook_func': 'authorizer_wsgi_entrance',
             'hook_module': 'tingyun.armoury.nova.api_ec2'},
            {"target": "nova.api.ec2", 'hook_func': 'executor_wsgi_entrance',
             'hook_module': 'tingyun.armoury.nova.api_ec2'},

            # webob.client
            {"target": "webob.client", 'hook_func': 'send_request_wsgi_entrance',
             'hook_module': 'tingyun.armoury.webob.client'},
            {"target": "webob.response", 'hook_func': 'response_wsgi_entrance',
             'hook_module': 'tingyun.armoury.webob.response'},
        ],
        # bottle 0.10.x-0.12.x
        "bottle": [
            {"target": "bottle", 'hook_func': 'detect_wsgi_entrance',
             'hook_module': 'tingyun.armoury.framework_bottle'},
            {"target": "bottle", 'hook_func': 'detect_templates',
             'hook_module': 'tingyun.armoury.framework_bottle'},
            {"target": "bottle", 'hook_func': 'detect_app_components',
             'hook_module': 'tingyun.armoury.framework_bottle'},
        ]
    }

    return hookers



"""this module used to wrap the specify method to RedisTrace

"""

from tingyun.armoury.ammunition.timer import Timer
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.logistics.warehouse.redis_node import RedisNode
from tingyun.logistics.basic_wrapper import wrap_object, FunctionWrapper


class RedisTrace(Timer):
    """
    """
    def __init__(self, tracker, command):
        """
        :return:
        """
        super(RedisTrace, self).__init__(tracker)
        self.command = command

    def create_node(self):
        """
        :return:
        """
        tracker = current_tracker()
        if tracker:
            tracker.redis_time = self.duration

        return RedisNode(command=self.command, children=self.children, start_time=self.start_time,
                         end_time=self.end_time, duration=self.duration, exclusive=self.exclusive)

    def terminal_node(self):
        return True


def redis_trace_wrapper(wrapped, command):
    """
    :return:
    """
    def dynamic_wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()
        if tracker is None:
            return wrapped(*args, **kwargs)

        if instance is not None:
            _command = command(instance, *args, **kwargs)
        else:
            _command = command(*args, **kwargs)

        with RedisTrace(tracker, _command):
            return wrapped(*args, **kwargs)

    def literal_wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()
        if tracker is None:
            return wrapped(*args, **kwargs)

        with RedisTrace(tracker, command):
            return wrapped(*args, **kwargs)

    if callable(command):
        return FunctionWrapper(wrapped, dynamic_wrapper)

    return FunctionWrapper(wrapped, literal_wrapper)


def wrap_redis_trace(module, object_path, command):
    wrap_object(module, object_path, redis_trace_wrapper, (command,))



import functools

from tingyun.logistics.basic_wrapper import FunctionWrapper, wrap_object
from tingyun.armoury.ammunition.tracker import current_tracker


class ErrorTrace(object):
    def __init__(self, tracker, ignore_errors=None):
        self._tracker = tracker
        self._ignore_errors = ignore_errors

    def __enter__(self):
        return self

    def __exit__(self, exc, value, tb):
        if exc is None or value is None or tb is None:
            return

        if self._tracker is None:
            return

        self._tracker.record_exception(exc=exc, value=value, tb=tb, ignore_errors=self._ignore_errors)


def error_trace_wrapper(wrapped, ignore_errors=None):

    def wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()

        if tracker is None:
            return wrapped(*args, **kwargs)

        with ErrorTrace(tracker, ignore_errors):
            return wrapped(*args, **kwargs)

    return FunctionWrapper(wrapped, wrapper)


def wrap_error_trace(module, object_path, ignore_errors=None):
    wrap_object(module, object_path, error_trace_wrapper, (ignore_errors, ))




"""this module is implement the function detector for flask

"""
import logging
from tingyun.logistics.basic_wrapper import FunctionWrapper
from tingyun.logistics.object_name import callable_name
from tingyun.armoury.ammunition.function_tracker import FunctionTracker
from tingyun.armoury.ammunition.tracker import current_tracker

console = logging.getLogger(__name__)


def _wrapper_deco(priority=None):
    def _wrapper(wrapped, instance, args, kwargs):
        """
        :param wrapped:
        :param instance:
        :param args:
        :param kwargs:
        :return:
        """
        tracker = current_tracker()
        if not tracker:
            return wrapped(*args, **kwargs)

        tracker.set_tracker_name(callable_name(wrapped), priority=priority)
        with FunctionTracker(tracker, callable_name(wrapped)):
            return wrapped(*args, **kwargs)

    return _wrapper


def add_url_rule_wrapper(wrapped, instance, args, kwargs):
    """used to trace the views metric
    :param wrapped:
    :param instance:
    :return:
    """
    def parse_view_func(rule, endpoint=None, view_func=None, **options):
        return rule, endpoint, view_func, options

    rule, endpoint, view_func, options = parse_view_func(*args, **kwargs)

    return wrapped(rule, endpoint, FunctionWrapper(view_func, _wrapper_deco(4)), **options)


def handle_exception_wrapper(wrapped, instance, args, kwargs):
    """used to trace the exception errors.
    :param wrapped:
    :param instance:
    :param args:
    :param kwargs:
    :return:
    """
    tracker = current_tracker()
    if not tracker:
        return wrapped(*args, **kwargs)

    # just record the exception info
    tracker.record_exception()

    return wrapped(*args, **kwargs)


def wrap_register_error(wrapped, instance, args, kwargs):
    """
    :param module:
    :return:
    """
    def parse_func_params(key, code_or_exception, f, *args, **kwargs):
        return key, code_or_exception, f, args, kwargs

    key, code_or_exception, f, _args, _kwargs = parse_func_params(*args, **kwargs)

    return wrapped(key, code_or_exception, FunctionWrapper(f, _wrapper_deco()), *_args, **_kwargs)


def wrap_func_with_request(wrapped, instance, args, kwargs):
    """
    :param wrapped:
    :param instance:
    :param args:
    :param kwargs:
    :return:
    """
    def parse_func_params(f, *args, **kwargs):
        return f, args, kwargs

    f, _args, _kwargs = parse_func_params(*args, **kwargs)
    return wrapped(FunctionWrapper(f, _wrapper_deco()), *_args, **_kwargs)


def wrap_template_filter(wrapped, instance, args, kwargs):
    """
    :param wrapped:
    :param instance:
    :param args:
    :param kwargs:
    :return:
    """
    def parse_func_params(name, *args, **kwargs):
        return name, args, kwargs

    name, _args, _kwargs = parse_func_params(*args, **kwargs)
    return FunctionWrapper(wrapped(name, *_args, **_kwargs), _wrapper_deco())




"""

"""

import logging

from tingyun.armoury.ammunition.timer import Timer
from tingyun.logistics.warehouse.external_node import ExternalNode
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.logistics.basic_wrapper import FunctionWrapper, wrap_object

console = logging.getLogger(__name__)


# defined error type for trace the external request errors.
MALFORMED_URL_ERROR_CODE = 900
UNKNOWN_HOST_ERROR_CODE = 901
CONNECT_EXCEPTION_ERROR_CODE = 902
TIMEOUT_EXCEPTION_ERROR_CODE = 903
CLIENT_PROTOCOL_ERROR_CODE = 904
CLIENT_ABORT_ERROR_CODE = 905
ILLEGAL_RESPONSE_ERROR_CODE = 906
SSL_EXCEPTION_ERROR_CODE = 908
OTHER_ERROR_CODE = 1000


class ExternalTrace(Timer):
    """define the external trace common api.

    """
    def __init__(self, tracker, library, url, params=None):
        super(ExternalTrace, self).__init__(tracker)

        self.library = library
        self.url = url.split('?')[0]
        self.params = parse_parameters(url, params)
        self.protocol = "http"
        self.protocol = "https" if "https" in url else self.protocol
        self.protocol = "thrift" if "thrift" in url else self.protocol

        signed_param = []
        for p in self.params:
            if tracker.settings and p in tracker.settings.external_url_params_captured.get(self.url, ""):
                signed_param.append("%s=%s&" % (p, self.params[p]))

        if 0 != len(signed_param):
            self.url = "%s?%s" % (self.url, ''.join(signed_param).rstrip('&'))

    def create_node(self):
        tracker = current_tracker()
        if tracker:
            tracker.external_time = self.duration

        return ExternalNode(library=self.library, url=self.url, children=self.children, protocol=self.protocol,
                            start_time=self.start_time, end_time=self.end_time, duration=self.duration,
                            exclusive=self.exclusive)

    def terminal_node(self):
        return True


def external_trace_wrapper(wrapped, library, url, params=None, exception_wrapper=None):
    """
    :param wrapped:
    :param library:
    :param url:
    :param params: just use for cross trace and error trace with replaced headers.
    :param exception_wrapper:
    :return:
    """

    def dynamic_wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()

        if tracker is None:
            return wrapped(*args, **kwargs)

        _url = url
        if callable(url):
            if instance is not None:
                _url = url(instance, *args, **kwargs)
            else:
                _url = url(*args, **kwargs)

        with ExternalTrace(tracker, library, _url, kwargs):
            if not callable(params):
                if callable(exception_wrapper):
                    return exception_wrapper(wrapped, _url, params, *args, **kwargs)

                return wrapped(*args, **kwargs)
            else:
                _args, _kwargs = params(*args, **kwargs)
                if callable(exception_wrapper):
                    ret = exception_wrapper(wrapped, _url, _kwargs.get("params"), *_args, **_kwargs)
                else:
                    ret = wrapped(*_args, **_kwargs)

                try:
                    # for requests/urllib3
                    if hasattr(ret, 'headers'):
                        tracker._called_traced_data = eval(ret.headers.get("X-Tingyun-Tx-Data", '{}'))

                    # for httplib2, note: the httplib2 will trans the upper case to lower case
                    if isinstance(ret, tuple):
                        tracker._called_traced_data = eval(ret[0].get("x-tingyun-tx-data", '{}'))
                except Exception as err:
                    console.debug(err)

                return ret

    def literal_wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()

        if tracker is None:
            return wrapped(*args, **kwargs)

        with ExternalTrace(tracker, library, url, kwargs):
            return wrapped(*args, **kwargs)

    if callable(url):
        return FunctionWrapper(wrapped, dynamic_wrapper)

    return FunctionWrapper(wrapped, literal_wrapper)


def wrap_external_trace(module, object_path, library, url, params=None, exception_wrapper=None):
    wrap_object(module, object_path, external_trace_wrapper, (library, url, params, exception_wrapper))


def process_cross_trace(headers):
    """
    """
    if headers is None:
        headers = {}

    tracker = current_tracker()
    if not tracker or not tracker.enabled:
        return headers

    if not tracker.settings.transaction_tracer.enabled:
        return headers

    headers['X-Tingyun-Id'] = tracker.settings.x_tingyun_id % (tracker._tingyun_id, tracker.generate_trace_id())

    return headers


def parse_parameters(_url, parameters):
    """
    :param _url: request url, which maybe contain get parameters
    :param parameters: keyword parameters
    :return:
    """
    if parameters and not isinstance(parameters, dict):
        return {}

    params_tmp = _url.split("?")
    _params = params_tmp[1] if 2 == len(params_tmp) else ""
    kv_params = [p for p in _params.split("&") if p.strip() and 2 == len(p.split("="))]

    actual_params = {}
    for kv in kv_params:
        k, v = kv.split('=', 1)
        if not k.strip() or not v.strip():
            continue

        actual_params[k.strip()] = v.strip()

    if parameters and isinstance(parameters, dict):
        actual_params.update(parameters)

    return actual_params




"""define some common method for tornado tracker
"""

import sys
import logging
import weakref
from tingyun.battlefield.tracer import Tracer
from tingyun.battlefield.knapsack import knapsack
from tingyun.battlefield.proxy import proxy_instance
from tingyun.logistics.object_name import callable_name
from tingyun.logistics.basic_wrapper import function_wrapper
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.ammunition.function_tracker import FunctionTracker

console = logging.getLogger(__name__)


def is_websocket(request):
    """Now web ignore websocket feature. it's not in our current conversation.
    :param request:
    :return:
    """
    if request.headers.get('Upgrade', '').lower() == 'websocket':
        console.info("current request is web socket, agent will ignore this tracker.")
        return True

    return False


class NoneProxy(object):
    pass


def record_exception(exc_info):
    """
    :return:
    """
    import tornado.web
    exc = exc_info[0]
    value = exc_info[1]

    # Not an error so we just return.
    if exc is tornado.web.Finish:
        return

    tracker = obtain_current_tracker()
    if tracker:
        tracker.record_exception(*exc_info)


def request_environ(environ, request):
    """the environ in the default environ may not include our need environment variable when apply to wsgi application
    :param environ: returned environment from the container request.
    :param request:
    :return:
    """
    # we need this the build the request information. but not included in the environment.
    ret = dict(environ)
    ret["REQUEST_URI"] = request.uri
    ret["HTTP_REFERER"] = request.headers.get('Referer', '')
    ret['HTTP_X_QUEUE_START'] = request.headers.get("X-Queue-Start")

    if not environ:
        ret['PATH_INFO'] = request.path
        ret['SCRIPT_NAME'] = ""
        ret['QUERY_STRING'] = request.query

    return ret


def obtain_current_tracker():
    """ Get the tracker from tracer cache.
    :return:
    """
    return current_tracker()


def obtain_tracker_from_request(request):
    """
    :param request: httpserver request
    :return:
    """
    tracker = getattr(request, '_self_tracker', None)
    if not tracker:
        console.debug("No tracker found in request[%s], if this continues, please report to us.", request)

    return tracker


def obtain_request(tracker):
    """obtain the request object from tracker. which is attached at began
    :param tracker:
    :return:
    """
    # its' weakref
    request = getattr(tracker, "_self_tracker", None)
    return request if not request else request()


def current_thread_id():
    return knapsack().current_thread_id()


def drop_current_tracer():
    """ separate the tracker from agent thread
    :return:
    """
    old_tracker = obtain_current_tracker()
    if old_tracker is not None:
        old_tracker.drop_tracker()
    return old_tracker


def finish_tracker(tracker, exc_type=None, exc_val=None, exc_tb=None):
    if not tracker:
        console.error("Tracker is None, shit maybe some trace strategy issue. if this continues. please report to us.")
        return

    if getattr(tracker, "_is_finalized"):
        console.error("This errors maybe caused by some agent logic trace error. if this continues, please report to"
                      " us for further investigation.")
        return

    # check the tracker status.
    if not (getattr(tracker, "_can_finalize") and getattr(tracker, "_ref_count") == 0):
        # console.debug("tracker can not finished. because of the some trace still in work.")
        return

    old_tracker = replace_current_tracer(tracker)
    try:
        tracker.finish_work(exc_type, exc_val, exc_tb, async=False)
    finally:
        setattr(tracker, "_is_finalized", True)
        request = obtain_request(tracker)
        if request:
            setattr(request, "_self_tracker", None)

        setattr(tracker, "_self_request", None)

        # put back the previous tracker to tracker cache.
        if tracker != old_tracker:
            replace_current_tracer(old_tracker)


def replace_current_tracer(new_tracker):
    """ Set the current tracer into agent thread and return the old one.
    :param new_tracker:
    :return:
    """
    old_tracker = drop_current_tracer()
    if new_tracker:
        new_tracker.save_tracker()

    return old_tracker


def generate_tracer(request, framework='Tornado'):
    """
    :param request: the http request for client
    :return:
    """
    drop_current_tracer()
    tracer = Tracer(proxy_instance(), request_environ({}, request), framework)
    if not tracer.enabled:
        return None

    tracer.start_work()
    drop_current_tracer()  # drop it from thread cache
    tracer._self_request = weakref.ref(request)

    # when 'finish' or 'on_connection_close' is called on server requests. tracer should be finished trace.
    tracer._is_finalized = False
    tracer._ref_count = 0  # count the reference.
    tracer._can_finalize = False

    return tracer


class TrackerTransferContext(object):
    def __init__(self, tracer):
        self.tracer = tracer

    def __enter__(self):
        self.old_tracer = replace_current_tracer(self.tracer)

    def __exit__(self, exc_type, exc_value, traceback):
        replace_current_tracer(self.old_tracer)


def create_tracker_aware_fxn(fxn, fxn_for_name=None, should_trace=True):
    if fxn is None or hasattr(fxn, '_self_tracker'):
        return None

    if fxn_for_name is None:
        fxn_for_name = fxn

    tracker = [obtain_current_tracker()]

    @function_wrapper
    def tracker_aware(wrapped, instance, args, kwargs):
        # Variables from the outer scope are not assignable in a closure, so we use a mutable object to hold the
        # tracker, so we can change it if we need to.
        inner_tracker = tracker[0]

        if inner_tracker is not None:
            # Callback run outside the main thread must not affect the cache
            if inner_tracker.thread_id != current_thread_id():
                return fxn(*args, **kwargs)

        if inner_tracker is not None and inner_tracker._is_finalized:
            inner_tracker = None
            tracker[0] = None

        with TrackerTransferContext(inner_tracker):
            if inner_tracker is None:
                # A tracker will be None for fxns scheduled on the ioloop not associated with a tracker.
                ret = fxn(*args, **kwargs)
            elif should_trace is False:
                try:
                    ret = fxn(*args, **kwargs)
                except:
                    record_exception(sys.exc_info())
                    wrapped._self_recorded_exception = True
                    raise

            else:
                name = callable_name(fxn_for_name)
                with FunctionTracker(inner_tracker, name=name) as ft:

                    try:
                        ret = fxn(*args, **kwargs)
                    except:
                        record_exception(sys.exc_info())
                        wrapped._self_recorded_exception = True
                        raise
                    if ft is not None and ret is not None and hasattr(ret, '_self_coroutine_name'):
                        ft.name = ret._self_coroutine_name

                        if type(ret) == NoneProxy:
                            ret = None

        if inner_tracker and inner_tracker._ref_count == 0:
            finish_tracker(inner_tracker)

        return ret

    return tracker_aware(fxn)



import cgi
import logging
import threading
import time

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from tingyun.armoury.ammunition.tracker import Tracker
from tingyun.packages import six


console = logging.getLogger(__name__)


class Tracer(Tracker):
    """
    """
    def __init__(self, application, environ, framework="Python"):
        """
        """
        Tracker.__init__(self, application, environ, framework)

        if not self.enabled:
            return

        self.get_thread_name()
        script_name = environ.get('SCRIPT_NAME', "")
        path_info = environ.get('PATH_INFO', "")
        self.request_uri = environ.get("REQUEST_URI", "")
        self.referer = environ.get("HTTP_REFERER", "")
        self.browser_header_prepared = False
        self.agent_browser_code = "ty_rum.agent={id:'%s', n:'%s', a:'%s', q:'%s', tid: '%s'}"
        self.http_method = environ.get('REQUEST_METHOD', "GET")

        if self.request_uri:
            self.request_uri = urlparse.urlparse(self.request_uri)[2]
            
        if script_name or path_info:
            if not path_info:
                path = script_name
            elif not script_name:
                path = path_info
            else:
                path = script_name + path_info

            self.set_tracker_name(path, 'Uri', priority=1)

            if not self.request_uri:
                self.request_uri = path
        else:
            if self.request_uri:
                self.set_tracker_name(self.request_uri, 'Uri', priority=1)

        # get the param
        self._get_request_params(environ)
        self.calculate_queque_time(environ)
        self.process_cross_trace(environ)

    def process_cross_trace(self, environ):
        """
        :return:
        """
        if not self.enabled:
            return

        # as caller
        tingyun_id = 'HTTP_X_TINGYUN_ID'
        tingyun_headers = environ.get(tingyun_id, "")

        if not tingyun_headers:
            return

        # not the same account,skip it
        ids = tingyun_headers.split(';')
        call_tingyun_id = ids[0].split('|')[0]
        called_tingyun_id = self._tingyun_id.split('|')[0]
        if call_tingyun_id != called_tingyun_id:
            self.call_tingyun_id = ""
            console.debug("Get the invalid tingyun id %s", environ.get(tingyun_id, ""))
            return

        id_dict = dict()
        for h in ids[1:]:
            if '=' not in h:
                continue

            kv = h.split('=')
            id_dict[kv[0]] = kv[1]

        self.call_tingyun_id = ids[0]
        self.call_req_id = id_dict.get("r", "")
        self._trace_id = id_dict.get("x", "")

    def deal_response(self, status, response_headers, *args):
        """
        :param status:
        :param response_headers:
        :param args:
        :return:
        """
        try:
            self.http_status = int(status.split(' ')[0])
        except Exception as _:
            console.warning("get status code failed, status is %s", status)

    def calculate_queque_time(self, env):
        """
        :return:
        """
        # for compitable with old version with modwsgi
        queue_header = ('HTTP_X_QUEUE_START', 'mod_wsgi.request_start', 'mod_wsgi.queue_start')
        x_queque = None

        for header in queue_header:
            x_queque = env.get(header, None)
            if x_queque:
                break

        if x_queque is None:
            return

        try:
            key, value = tuple(x_queque.split('='))
            # deal value from nginx
            if 't' == key:
                self.queque_start = float(value) / 1000.0
            else:
                self.queque_start = float(value) * 1000.0
        except Exception as err:
            pass

    def _get_request_params(self, environ):
        """
        :param environ: cgi environment
        """
        value = environ.get('QUERY_STRING', "")
        if value:
            params = {}

            try:
                params = urlparse.parse_qs(value, keep_blank_values=True)
            except Exception:
                try:
                    # method <parse_qs> only for backward compatibility,  method <parse_qs> new in python2.6
                    params = cgi.parse_qs(value, keep_blank_values=True)
                except Exception:
                    pass

            for k, v in params.items():
                if 1 == len(v):
                    params[k] = v[0]

            self.request_params.update(params)

        # avoiding wasting cpu when customize web naming closed
        if not self.settings.naming.should_naming:
            return

        cookie_params = environ.get('HTTP_COOKIE', '').split(";")
        for kv in cookie_params:
            if '=' not in kv:
                continue

            k, v = kv.split('=')
            self.cookie_params[k.strip()] = v.strip()

        for p in environ:
            if not p.startswith('HTTP_'):
                continue

            k = p[5:].replace('_', '-').lower()
            self.header_params[k] = environ.get(p)

    def get_thread_name(self):
        """
        :return: thread name
        """
        try:
            self.thread_name = threading.currentThread().getName()
        except Exception as err:
            console.info("Get thread name failed. %s", err)
            self.thread_name = "unknown"

        return self.thread_name

    def start_work(self):
        """
        :return:
        """
        self.__enter__()

    def finish_work(self, exc_type, exc_val, exc_tb, async=False):
        """
        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        self.__exit__(exc_type, exc_val, exc_tb, async)

    def generate_agent_param(self):
        """
        :return:
        """
        duration = int((time.time() - self.start_time) * 1000)
        queque_time = self.queque_time
        guid = ""

        if self.settings and self.settings.action_tracer.action_threshold < duration:
            guid = self.generate_trace_guid()

        bids = self._settings.tingyunIdSecret.split('|')
        bid = bids[1] if len(bids) == 2 else "do-not-get-id-for-callable"
        agent_param = self.agent_browser_code % (bid, self.path, duration, queque_time, guid)

        return agent_param

    def fork_browser_js_head(self):
        """generate the head for rum
        :return:
        """
        header = b''

        if not self.settings or not self.enabled:
            return header

        if not self.settings.rum.enabled:
            return header

        if self.browser_header_prepared:
            return header

        if not self.settings.rum.script:
            return header

        agent_param = self.generate_agent_param()
        param_index = str(self.settings.rum.script).rfind('}')
        header = "<script type='text/javascript'>%s%s%s</script>" % (self.settings.rum.script[:param_index],
                                                                     agent_param,
                                                                     self.settings.rum.script[param_index:])

        # avoid some issues, just use the ascii encoding to specify the insert js script in head tag
        try:
            if six.PY2:
                header = header.encode('ascii')
            else:
                header.encode('ascii')
        except UnicodeError:
            header = b''

        self.browser_header_prepared = True if header else False

        return header




"""all the function in the battle traced will use this function tracker.

"""
import logging
import functools
import traceback

from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.logistics.warehouse.function_node import FunctionNode
from tingyun.armoury.ammunition.timer import Timer
from tingyun.logistics.basic_wrapper import FunctionWrapper, wrap_object
from tingyun.logistics.object_name import callable_name

console = logging.getLogger(__name__)


class FunctionTracker(Timer):
    """
    """
    def __init__(self, tracker, name, group=None, label=None, params=None):
        super(FunctionTracker, self).__init__(tracker)

        self.name = name
        self.group = group or 'Function'
        self.label = label
        self.params = params

        self.stack_trace = None

    def create_node(self):
        """create function trace node
        """
        return FunctionNode(group=self.group, name=self.name, children=self.children, start_time=self.start_time,
                            end_time=self.end_time, duration=self.duration, exclusive=self.exclusive,
                            params=self.params, stack_trace=self.stack_trace)

    def finalize_data(self):
        """create all the data if need
        :return:
        """
        settings = self.tracker.settings

        if settings.action_tracer.enabled and self.duration >= settings.action_tracer.stack_trace_threshold:
            if self.tracker.stack_trace_count < settings.stack_trace_count:
                self.stack_trace = traceback.extract_stack()
                self.tracker.stack_trace_count += 1


def function_trace_wrapper(wrapped, name=None, group=None, label=None, params=None):

    def dynamic_wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()

        if tracker is None:
            return wrapped(*args, **kwargs)

        _name = name
        _params = params
        _label = label
        _group = group

        if callable(name):
            if instance is not None:
                _name = name(instance, *args, **kwargs)
            else:
                _name = name(*args, **kwargs)
        elif name is None:
            _name = callable_name(wrapped)

        if callable(group):
            if instance is not None:
                _group = group(instance, *args, **kwargs)
            else:
                _group = group(*args, **kwargs)

        with FunctionTracker(tracker, _name, _group, _label, _params):
            return wrapped(*args, **kwargs)

    def literal_wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()

        if tracker is None:
            return wrapped(*args, **kwargs)

        _name = name or callable_name(wrapped)

        with FunctionTracker(tracker, _name, group, label, params):
            return wrapped(*args, **kwargs)

    if callable(name) or callable(group) or callable(label) or callable(params):
        return FunctionWrapper(wrapped, dynamic_wrapper)

    return FunctionWrapper(wrapped, literal_wrapper)


def wrap_function_trace(module, object_path, name=None, group=None, label=None, params=None):
    return wrap_object(module, object_path, function_trace_wrapper, (name, group, label, params))


def function_trace_decorator(name=None, group=None):
    """
    :param name: the name of the function named to
    :param group: the group of the function belong to. default `Function`
    :return:
    """
    return functools.partial(function_trace_wrapper, name=name, group=group)



"""this module used to wrap the specify armory to trace

"""

import logging

from tingyun.armoury.ammunition.timer import Timer
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.logistics.warehouse.mongo_node import MongoNode
from tingyun.logistics.basic_wrapper import wrap_object, FunctionWrapper


console = logging.getLogger(__name__)


class MongoTracker(Timer):
    """
    """
    def __init__(self, tracker, schema, method):
        """
        :return:
        """
        super(MongoTracker, self).__init__(tracker)
        self.method = method
        self.schema = schema

    def create_node(self):
        """
        :return:
        """
        tracker = current_tracker()
        if tracker:
            tracker.mongo_time = self.duration

        return MongoNode(method=self.method, children=self.children, start_time=self.start_time, schema=self.schema,
                         end_time=self.end_time, duration=self.duration, exclusive=self.exclusive)

    def terminal_node(self):
        return True


def mongo_trace_wrapper(wrapped, schema, method):
    """
    :return:
    """
    def dynamic_wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()
        if tracker is None:
            return wrapped(*args, **kwargs)

        _method = method
        if callable(method):
            if instance is not None:
                _method = method(instance, *args, **kwargs)
            else:
                _method = method(*args, **kwargs)

        _schema = schema
        if callable(schema):
            if instance is not None:
                _schema = schema(instance, *args, **kwargs)
            else:
                _schema = schema(*args, **kwargs)

        with MongoTracker(tracker, _schema, _method):
            return wrapped(*args, **kwargs)

    def literal_wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()
        if tracker is None:
            return wrapped(*args, **kwargs)

        with MongoTracker(tracker, method):
            return wrapped(*args, **kwargs)

    if callable(method) or callable(schema):
        return FunctionWrapper(wrapped, dynamic_wrapper)

    return FunctionWrapper(wrapped, literal_wrapper)


def wrap_mongo_trace(module, object_path, schema, method):
    wrap_object(module, object_path, mongo_trace_wrapper, (schema, method,))
